# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from datetime import datetime, timedelta
from typing import Final, Literal, Optional, Union

import pandas as pd
import taosws

import mlrun.common.schemas.model_monitoring as mm_schemas
import mlrun.common.types
import mlrun.model_monitoring.db.tsdb.tdengine.schemas as tdengine_schemas
from mlrun.config import config
from mlrun.datastore.datastore_profile import DatastoreProfile
from mlrun.model_monitoring.db import TSDBConnector
from mlrun.model_monitoring.db.tsdb.tdengine.tdengine_connection import (
    Statement,
    TDEngineConnection,
)
from mlrun.model_monitoring.helpers import get_invocations_fqn, get_start_end
from mlrun.utils import logger

# Thread-local storage for connections
_thread_local = threading.local()


class TDEngineTimestampPrecision(mlrun.common.types.StrEnum):
    """
    The timestamp precision for the TDEngine database.
    For more information, see:
    https://docs.tdengine.com/tdengine-reference/sql-manual/data-types/#timestamp
    https://docs.tdengine.com/tdengine-reference/sql-manual/manage-databases/#create-database
    """

    MILLISECOND = "ms"  # TDEngine's default
    MICROSECOND = "us"  # MLRun's default
    NANOSECOND = "ns"


class TDEngineConnector(TSDBConnector):
    """
    Handles the TSDB operations when the TSDB connector is of type TDEngine.
    """

    type: str = mm_schemas.TSDBTarget.TDEngine

    def __init__(
        self,
        project: str,
        profile: DatastoreProfile,
        timestamp_precision: TDEngineTimestampPrecision = TDEngineTimestampPrecision.MICROSECOND,
    ):
        super().__init__(project=project)

        self._tdengine_connection_profile = profile

        self._timestamp_precision: Final = (  # cannot be changed after initialization
            timestamp_precision
        )

        if not mlrun.mlconf.system_id:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "system_id is not set in mlrun.mlconf. "
                "TDEngineConnector requires system_id to be configured for database name construction. "
                "Please ensure MLRun configuration is properly loaded before creating TDEngineConnector."
            )
        self.database = (
            f"{tdengine_schemas._MODEL_MONITORING_DATABASE}_{mlrun.mlconf.system_id}"
        )
        self._init_super_tables()

    @property
    def connection(self) -> TDEngineConnection:
        if not hasattr(_thread_local, "connection"):
            _thread_local.connection = self._create_connection()
            logger.debug(
                "Created new TDEngine connection for thread",
                project=self.project,
                thread_name=threading.current_thread().name,
                thread_id=threading.get_ident(),
            )
        return _thread_local.connection

    def _create_connection(self) -> TDEngineConnection:
        """Establish a connection to the TSDB server."""
        logger.debug("Creating a new connection to TDEngine", project=self.project)
        conn = TDEngineConnection(
            self._tdengine_connection_profile.dsn(),
        )
        conn.prefix_statements = [f"USE {self.database}"]

        return conn

    def _init_super_tables(self):
        """Initialize the super tables for the TSDB."""
        self.tables = {
            mm_schemas.TDEngineSuperTables.APP_RESULTS: tdengine_schemas.AppResultTable(
                project=self.project, database=self.database
            ),
            mm_schemas.TDEngineSuperTables.METRICS: tdengine_schemas.Metrics(
                project=self.project, database=self.database
            ),
            mm_schemas.TDEngineSuperTables.PREDICTIONS: tdengine_schemas.Predictions(
                project=self.project, database=self.database
            ),
            mm_schemas.TDEngineSuperTables.ERRORS: tdengine_schemas.Errors(
                project=self.project, database=self.database
            ),
        }

    def _create_db_if_not_exists(self):
        """Create the database if it does not exist."""
        self.connection.prefix_statements = []
        self.connection.run(
            statements=f"CREATE DATABASE IF NOT EXISTS {self.database} PRECISION '{self._timestamp_precision}'",
        )
        self.connection.prefix_statements = [f"USE {self.database}"]
        logger.debug(
            "The TDEngine database is currently in use",
            project=self.project,
            database=self.database,
        )

    def create_tables(self):
        """Create TDEngine supertables."""

        # Create the database if it does not exist
        self._create_db_if_not_exists()

        for table in self.tables:
            create_table_query = self.tables[table]._create_super_table_query()
            conn = self.connection
            conn.run(
                statements=create_table_query,
            )

    def write_application_event(
        self,
        event: dict,
        kind: mm_schemas.WriterEventKind = mm_schemas.WriterEventKind.RESULT,
    ) -> None:
        """
        Write a single result or metric to TSDB.
        """

        table_name = (
            f"{event[mm_schemas.WriterEvent.ENDPOINT_ID]}_"
            f"{event[mm_schemas.WriterEvent.APPLICATION_NAME]}"
        )

        if kind == mm_schemas.WriterEventKind.RESULT:
            # Write a new result
            table = self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS]
            table_name = (
                f"{table_name}_{event[mm_schemas.ResultData.RESULT_NAME]}"
            ).replace("-", "_")

        else:
            # Write a new metric
            table = self.tables[mm_schemas.TDEngineSuperTables.METRICS]
            table_name = (
                f"{table_name}_{event[mm_schemas.MetricData.METRIC_NAME]}"
            ).replace("-", "_")

        # Escape the table name for case-sensitivity (ML-7908)
        # https://github.com/taosdata/taos-connector-python/issues/260
        table_name = f"`{table_name}`"

        # Convert the datetime strings to datetime objects
        event[mm_schemas.WriterEvent.END_INFER_TIME] = self._convert_to_datetime(
            val=event[mm_schemas.WriterEvent.END_INFER_TIME]
        )
        event[mm_schemas.WriterEvent.START_INFER_TIME] = self._convert_to_datetime(
            val=event[mm_schemas.WriterEvent.START_INFER_TIME]
        )

        create_table_sql = table._create_subtable_sql(subtable=table_name, values=event)

        # we need the string values to be sent to the connection, not the enum
        columns = {str(key): str(val) for key, val in table.columns.items()}

        insert_statement = Statement(
            columns=columns,
            subtable=table_name,
            values=event,
            timestamp_precision=self._timestamp_precision,
        )

        self.connection.run(
            statements=[
                create_table_sql,
                insert_statement,
            ],
        )

    @staticmethod
    def _convert_to_datetime(val: Union[str, datetime]) -> datetime:
        return datetime.fromisoformat(val) if isinstance(val, str) else val

    @staticmethod
    def _generate_filter_query(
        filter_column: str, filter_values: Union[str, list[Union[str, int]]]
    ) -> str:
        """
        Generate a filter query for TDEngine based on the provided column and values.

        :param filter_column: The column to filter by.
        :param filter_values: A single value or a list of values to filter by.

        :return: A string representing the filter query.
        :raise: ``MLRunValueError`` if the filter values are not of type string or list.
        """
        if isinstance(filter_values, str):
            return f"{filter_column}='{filter_values}'"
        elif isinstance(filter_values, list):
            return f"{filter_column} IN ({', '.join(repr(v) for v in filter_values)}) "
        else:
            raise mlrun.errors.MLRunValueError(
                f"Invalid filter values {filter_values}: must be a string or a list, "
                f"got {type(filter_values).__name__}; filter values: {filter_values}"
            )

    def _drop_database_query(self) -> str:
        return f"DROP DATABASE IF EXISTS {self.database};"

    def _get_table_name_query(self) -> str:
        return f"SELECT table_name FROM information_schema.ins_tables where db_name='{self.database}' LIMIT 1;"

    def apply_monitoring_stream_steps(self, graph, **kwarg):
        """
        Apply TSDB steps on the provided monitoring graph. Throughout these steps, the graph stores live data of
        different key metric dictionaries. This data is being used by the monitoring dashboards in
        grafana. At the moment, we store two types of data:
        - prediction latency.
        - custom metrics.
        """

        def apply_process_before_tsdb():
            graph.add_step(
                "mlrun.model_monitoring.db.tsdb.tdengine.stream_graph_steps.ProcessBeforeTDEngine",
                name="ProcessBeforeTDEngine",
                after="FilterNOP",
            )

        def apply_tdengine_target(name, after):
            graph.add_step(
                "mlrun.datastore.storeytargets.TDEngineStoreyTarget",
                name=name,
                after=after,
                url=f"ds://{self._tdengine_connection_profile.name}",
                supertable=self.tables[
                    mm_schemas.TDEngineSuperTables.PREDICTIONS
                ].super_table,
                table_col=mm_schemas.EventFieldType.TABLE_COLUMN,
                time_col=mm_schemas.EventFieldType.TIME,
                database=self.database,
                columns=[
                    mm_schemas.EventFieldType.LATENCY,
                    mm_schemas.EventKeyMetrics.CUSTOM_METRICS,
                    mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT,
                    mm_schemas.EventFieldType.EFFECTIVE_SAMPLE_COUNT,
                ],
                tag_cols=[
                    mm_schemas.EventFieldType.ENDPOINT_ID,
                ],
                max_events=1000,
                flush_after_seconds=30,
            )

        apply_process_before_tsdb()
        apply_tdengine_target(
            name="TDEngineTarget",
            after="ProcessBeforeTDEngine",
        )

    def add_pre_writer_steps(self, graph, after):
        return graph.add_step(
            "mlrun.model_monitoring.db.tsdb.tdengine.writer_graph_steps.ProcessBeforeTDEngine",
            name="ProcessBeforeTDEngine",
            after=after,
        )

    def apply_writer_steps(self, graph, after, **kwargs) -> None:
        graph.add_step(
            "mlrun.datastore.storeytargets.TDEngineStoreyTarget",
            name="tsdb_metrics",
            after=after,
            url=f"ds://{self._tdengine_connection_profile.name}",
            supertable=self.tables[mm_schemas.TDEngineSuperTables.METRICS].super_table,
            table_col=mm_schemas.EventFieldType.TABLE_COLUMN,
            time_col=mm_schemas.WriterEvent.END_INFER_TIME,
            database=self.database,
            graph_shape="cylinder",
            columns=[
                mm_schemas.WriterEvent.START_INFER_TIME,
                mm_schemas.MetricData.METRIC_VALUE,
            ],
            tag_cols=[
                mm_schemas.WriterEvent.ENDPOINT_ID,
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.MetricData.METRIC_NAME,
            ],
            max_events=config.model_endpoint_monitoring.writer_graph.max_events,
            flush_after_seconds=config.model_endpoint_monitoring.writer_graph.flush_after_seconds,
        )

        graph.add_step(
            "mlrun.datastore.storeytargets.TDEngineStoreyTarget",
            name="tsdb_app_results",
            after=after,
            url=f"ds://{self._tdengine_connection_profile.name}",
            supertable=self.tables[
                mm_schemas.TDEngineSuperTables.APP_RESULTS
            ].super_table,
            table_col=mm_schemas.EventFieldType.TABLE_COLUMN,
            time_col=mm_schemas.WriterEvent.END_INFER_TIME,
            database=self.database,
            graph_shape="cylinder",
            columns=[
                mm_schemas.WriterEvent.START_INFER_TIME,
                mm_schemas.ResultData.RESULT_VALUE,
                mm_schemas.ResultData.RESULT_STATUS,
                mm_schemas.ResultData.RESULT_EXTRA_DATA,
            ],
            tag_cols=[
                mm_schemas.WriterEvent.ENDPOINT_ID,
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.ResultData.RESULT_NAME,
                mm_schemas.ResultData.RESULT_KIND,
            ],
            max_events=config.model_endpoint_monitoring.writer_graph.max_events,
            flush_after_seconds=config.model_endpoint_monitoring.writer_graph.flush_after_seconds,
        )

    def handle_model_error(
        self,
        graph,
        tsdb_batching_max_events: int = 1000,
        tsdb_batching_timeout_secs: int = 30,
        **kwargs,
    ) -> None:
        graph.add_step(
            "mlrun.model_monitoring.db.tsdb.tdengine.stream_graph_steps.ErrorExtractor",
            name="error_extractor",
            after="ForwardError",
        )
        graph.add_step(
            "mlrun.datastore.storeytargets.TDEngineStoreyTarget",
            name="tsdb_error",
            after="error_extractor",
            url=f"ds://{self._tdengine_connection_profile.name}",
            supertable=self.tables[mm_schemas.TDEngineSuperTables.ERRORS].super_table,
            table_col=mm_schemas.EventFieldType.TABLE_COLUMN,
            time_col=mm_schemas.EventFieldType.TIME,
            database=self.database,
            columns=[
                mm_schemas.EventFieldType.MODEL_ERROR,
            ],
            tag_cols=[
                mm_schemas.EventFieldType.ENDPOINT_ID,
                mm_schemas.EventFieldType.ERROR_TYPE,
            ],
            max_events=tsdb_batching_max_events,
            flush_after_seconds=tsdb_batching_timeout_secs,
        )

    def delete_tsdb_records(self, endpoint_ids: list[str]) -> None:
        """
        To delete subtables within TDEngine, we first query the subtables names with the provided endpoint_ids.
        Then, we drop each subtable.
        """
        logger.debug(
            "Deleting model endpoint resources using the TDEngine connector",
            project=self.project,
            number_of_endpoints_to_delete=len(endpoint_ids),
        )

        # Get all subtables with the provided endpoint_ids
        subtables = []
        try:
            for table in self.tables:
                get_subtable_query = self.tables[table]._get_subtables_query_by_tag(
                    filter_tag="endpoint_id", filter_values=endpoint_ids
                )
                subtables_result = self.connection.run(query=get_subtable_query)
                subtables.extend([subtable[0] for subtable in subtables_result.data])
        except Exception as e:
            logger.warning(
                "Failed to get subtables for deletion. You may need to delete them manually."
                "These can be found under the following supertables: app_results, "
                "metrics, errors, and predictions.",
                project=self.project,
                error=mlrun.errors.err_to_str(e),
            )

        # Prepare the drop statements
        drop_statements = [
            self.tables[table].drop_subtable_query(subtable=subtable)
            for subtable in subtables
        ]
        try:
            logger.debug("Dropping subtables", drop_statements=drop_statements)
            self.connection.run(statements=drop_statements)
        except Exception as e:
            logger.warning(
                "Failed to delete model endpoint resources. You may need to delete them manually. "
                "These can be found under the following supertables: app_results, "
                "metrics, errors, and predictions.",
                project=self.project,
                error=mlrun.errors.err_to_str(e),
            )
        logger.debug(
            "Deleted all model endpoint resources using the TDEngine connector",
            project=self.project,
            number_of_endpoints_to_delete=len(endpoint_ids),
        )

    def delete_application_records(
        self, application_name: str, endpoint_ids: Optional[list[str]] = None
    ) -> None:
        """
        Delete application records from the TSDB for the given model endpoints or all if ``endpoint_ids`` is ``None``.
        """
        logger.debug(
            "Deleting application records",
            project=self.project,
            application_name=application_name,
            endpoint_ids=endpoint_ids,
        )
        tables = [
            self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS],
            self.tables[mm_schemas.TDEngineSuperTables.METRICS],
        ]

        filter_query = self._generate_filter_query(
            filter_column=mm_schemas.ApplicationEvent.APPLICATION_NAME,
            filter_values=application_name,
        )
        if endpoint_ids:
            endpoint_ids_filter = self._generate_filter_query(
                filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
                filter_values=endpoint_ids,
            )
            filter_query += f" AND {endpoint_ids_filter}"

        drop_statements: list[str] = []
        for table in tables:
            get_subtable_query = table._get_tables_query_by_condition(filter_query)
            subtables_result = self.connection.run(query=get_subtable_query)
            drop_statements.extend(
                [
                    table.drop_subtable_query(subtable=subtable[0])
                    for subtable in subtables_result.data
                ]
            )

        logger.debug("Dropping application records", drop_statements=drop_statements)
        self.connection.run(statements=drop_statements)

    def delete_tsdb_resources(self):
        """
        Delete all project resources in the TSDB connector, such as model endpoints data and drift results.
        """
        logger.debug(
            "Deleting all project resources using the TDEngine connector",
            project=self.project,
        )
        drop_statements = []
        for table in self.tables:
            drop_statements.append(self.tables[table].drop_supertable_query())

        try:
            self.connection.run(
                statements=drop_statements,
            )
        except Exception as e:
            logger.warning(
                "Failed to drop TDEngine tables. You may need to drop them manually. "
                "These can be found under the following supertables: app_results, "
                "metrics, errors, and predictions.",
                project=self.project,
                error=mlrun.errors.err_to_str(e),
            )
        logger.debug(
            "Deleted all project resources using the TDEngine connector",
            project=self.project,
        )

        # Check if database is empty and if so, drop it
        self._drop_database_if_empty()

    def _drop_database_if_empty(self):
        query_random_table_name = self._get_table_name_query()
        drop_database = False
        try:
            table_name = self.connection.run(
                query=query_random_table_name,
            )
            if len(table_name.data) == 0:
                # no tables were found under the database
                drop_database = True

        except Exception as e:
            logger.warning(
                "Failed to query tables in the database. You may need to drop the database manually if it is empty.",
                project=self.project,
                error=mlrun.errors.err_to_str(e),
            )

        if drop_database:
            logger.debug(
                "Going to drop the TDEngine database",
                project=self.project,
                database=self.database,
            )
            drop_database_query = self._drop_database_query()
            try:
                self.connection.run(
                    statements=drop_database_query,
                )
                logger.debug(
                    "The TDEngine database has been successfully dropped",
                    project=self.project,
                    database=self.database,
                )

            except Exception as e:
                logger.warning(
                    "Failed to drop the database. You may need to drop it manually if it is empty.",
                    project=self.project,
                    error=mlrun.errors.err_to_str(e),
                )

    def get_model_endpoint_real_time_metrics(
        self,
        endpoint_id: str,
        metrics: list[str],
        start: str,
        end: str,
    ) -> dict[str, list[tuple[str, float]]]:
        # Not implemented, use get_records() instead
        pass

    def _get_records(
        self,
        table: str,
        start: datetime,
        end: datetime,
        columns: Optional[list[str]] = None,
        filter_query: Optional[str] = None,
        interval: Optional[str] = None,
        agg_funcs: Optional[list] = None,
        limit: Optional[int] = None,
        sliding_window_step: Optional[str] = None,
        timestamp_column: str = mm_schemas.EventFieldType.TIME,
        group_by: Optional[Union[list[str], str]] = None,
        preform_agg_columns: Optional[list] = None,
        order_by: Optional[str] = None,
        desc: Optional[bool] = None,
        partition_by: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Getting records from TSDB data collection.
        :param table:                 Either a supertable or a subtable name.
        :param start:                 The start time of the metrics.
        :param end:                   The end time of the metrics.
        :param columns:               Columns to include in the result.
        :param filter_query:          Optional filter expression as a string. TDengine supports SQL-like syntax.
        :param interval:              The interval to aggregate the data by. Note that if interval is provided,
                                      `agg_funcs` must bg provided as well. Provided as a string in the format of '1m',
                                      '1h', etc.
        :param agg_funcs:             The aggregation functions to apply on the columns. Note that if `agg_funcs` is
                                      provided, `interval` must bg provided as well. Provided as a list of strings in
                                      the format of ['sum', 'avg', 'count', ...].
        :param limit:                 The maximum number of records to return.
        :param sliding_window_step:   The time step for which the time window moves forward. Note that if
                                      `sliding_window_step` is provided, interval must be provided as well. Provided
                                      as a string in the format of '1m', '1h', etc.
        :param timestamp_column:      The column name that holds the timestamp index.
        :param group_by:              The column name to group by. Note that if `group_by` is provided, aggregation
                                      functions must bg provided
        :param preform_agg_columns:   The columns to preform aggregation on.
                                      notice that all aggregation functions provided will preform on those columns.
                                      If not provided The default behavior is to preform on all columns in columns,
                                      if an empty list was provided The aggregation won't be performed.
        :param order_by:              The column or alias to preform ordering on the query.
        :param desc:                  Whether or not to sort the results in descending order.
        :param partition_by:          The column to partition the results by. Note that if interval is provided,
                                      `agg_funcs` must bg provided as well.

        :return: DataFrame with the provided attributes from the data collection.
        :raise:  MLRunInvalidArgumentError if query the provided table failed.
        """

        full_query = tdengine_schemas.TDEngineSchema._get_records_query(
            table=table,
            start=start,
            end=end,
            columns_to_filter=columns,
            filter_query=filter_query,
            interval=interval,
            limit=limit,
            agg_funcs=agg_funcs,
            sliding_window_step=sliding_window_step,
            timestamp_column=timestamp_column,
            database=self.database,
            group_by=group_by,
            preform_agg_funcs_columns=preform_agg_columns,
            order_by=order_by,
            desc=desc,
            partition_by=partition_by,
        )
        logger.debug("Querying TDEngine", query=full_query)
        try:
            query_result = self.connection.run(
                query=full_query,
            )
        except taosws.QueryError as e:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Failed to query table {table} in database {self.database}, {str(e)}"
            )

        df_columns = [field.name for field in query_result.fields]
        return pd.DataFrame(query_result.data, columns=df_columns)

    def read_metrics_data(
        self,
        *,
        endpoint_id: str,
        start: datetime,
        end: datetime,
        metrics: list[mm_schemas.ModelEndpointMonitoringMetric],
        type: Literal["metrics", "results"],
        with_result_extra_data: bool = False,
    ) -> Union[
        list[
            Union[
                mm_schemas.ModelEndpointMonitoringResultValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ],
        ],
        list[
            Union[
                mm_schemas.ModelEndpointMonitoringMetricValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ],
        ],
    ]:
        timestamp_column = mm_schemas.WriterEvent.END_INFER_TIME
        columns = [timestamp_column, mm_schemas.WriterEvent.APPLICATION_NAME]
        if type == "metrics":
            if with_result_extra_data:
                logger.warning(
                    "The 'with_result_extra_data' parameter is not supported for metrics, just for results",
                    project=self.project,
                    endpoint_id=endpoint_id,
                )
            table = self.tables[mm_schemas.TDEngineSuperTables.METRICS].super_table
            name = mm_schemas.MetricData.METRIC_NAME
            columns += [name, mm_schemas.MetricData.METRIC_VALUE]
            df_handler = self.df_to_metrics_values
        elif type == "results":
            table = self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS].super_table
            name = mm_schemas.ResultData.RESULT_NAME
            columns += [
                name,
                mm_schemas.ResultData.RESULT_VALUE,
                mm_schemas.ResultData.RESULT_STATUS,
                mm_schemas.ResultData.RESULT_KIND,
            ]
            if with_result_extra_data:
                columns.append(mm_schemas.ResultData.RESULT_EXTRA_DATA)
            df_handler = self.df_to_results_values
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Invalid type {type}, must be either 'metrics' or 'results'."
            )

        metrics_condition = " OR ".join(
            [
                f"({mm_schemas.WriterEvent.APPLICATION_NAME}='{metric.app}' AND {name}='{metric.name}')"
                for metric in metrics
            ]
        )
        filter_query = f"(endpoint_id='{endpoint_id}') AND ({metrics_condition})"

        df = self._get_records(
            table=table,
            start=start,
            end=end,
            filter_query=filter_query,
            timestamp_column=timestamp_column,
            columns=columns,
        )

        df[mm_schemas.WriterEvent.END_INFER_TIME] = pd.to_datetime(
            df[mm_schemas.WriterEvent.END_INFER_TIME]
        )
        df.set_index(mm_schemas.WriterEvent.END_INFER_TIME, inplace=True)

        logger.debug(
            "Converting a DataFrame to a list of metrics or results values",
            table=table,
            project=self.project,
            endpoint_id=endpoint_id,
            is_empty=df.empty,
        )

        if not with_result_extra_data and type == "results":
            # Set the extra data to an empty string if it's not requested
            df[mm_schemas.ResultData.RESULT_EXTRA_DATA] = ""

        return df_handler(df=df, metrics=metrics, project=self.project)

    def read_predictions(
        self,
        *,
        endpoint_id: str,
        start: datetime,
        end: datetime,
        aggregation_window: Optional[str] = None,
        agg_funcs: Optional[list] = None,
        limit: Optional[int] = None,
    ) -> Union[
        mm_schemas.ModelEndpointMonitoringMetricValues,
        mm_schemas.ModelEndpointMonitoringMetricNoData,
    ]:
        if (agg_funcs and not aggregation_window) or (
            aggregation_window and not agg_funcs
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "both or neither of `aggregation_window` and `agg_funcs` must be provided"
            )
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.PREDICTIONS].super_table,
            start=start,
            end=end,
            columns=[mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT],
            filter_query=f"endpoint_id='{endpoint_id}'",
            agg_funcs=agg_funcs,
            interval=aggregation_window,
            limit=limit,
        )

        full_name = get_invocations_fqn(self.project)

        if df.empty:
            return mm_schemas.ModelEndpointMonitoringMetricNoData(
                full_name=full_name,
                type=mm_schemas.ModelEndpointMonitoringMetricType.METRIC,
            )

        if aggregation_window:
            # _wend column, which represents the end time of each window, will be used as the time index
            df["_wend"] = pd.to_datetime(df["_wend"])
            df.set_index("_wend", inplace=True)

        estimated_prediction_count = (
            f"{agg_funcs[0]}({mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT})"
            if agg_funcs
            else mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT
        )

        return mm_schemas.ModelEndpointMonitoringMetricValues(
            full_name=full_name,
            values=list(
                zip(
                    df.index,
                    df[estimated_prediction_count],
                )
            ),  # pyright: ignore[reportArgumentType]
        )

    def get_last_request(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> Union[pd.DataFrame, dict[str, float]]:
        if not endpoint_ids:
            return {}
        filter_query = self._generate_filter_query(
            filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
            filter_values=endpoint_ids,
        )
        start, end = get_start_end(start, end)
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.PREDICTIONS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.EventFieldType.ENDPOINT_ID,
                mm_schemas.EventFieldType.TIME,
                mm_schemas.EventFieldType.LATENCY,
            ],
            filter_query=filter_query,
            timestamp_column=mm_schemas.EventFieldType.TIME,
            agg_funcs=["last"],
            group_by=mm_schemas.EventFieldType.ENDPOINT_ID,
            preform_agg_columns=[mm_schemas.EventFieldType.TIME],
        )
        if not df.empty:
            df.dropna(inplace=True)
        df.rename(
            columns={
                f"last({mm_schemas.EventFieldType.TIME})": mm_schemas.EventFieldType.LAST_REQUEST,
                f"{mm_schemas.EventFieldType.LATENCY}": "last_latency",
            },
            inplace=True,
        )
        df[mm_schemas.EventFieldType.LAST_REQUEST] = pd.to_datetime(
            df[mm_schemas.EventFieldType.LAST_REQUEST],
            errors="coerce",
            format="ISO8601",
            utc=True,
        )
        return df

    def get_drift_status(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> pd.DataFrame:
        filter_query = self._generate_filter_query(
            filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
            filter_values=endpoint_ids,
        )
        start = start or (mlrun.utils.datetime_now() - timedelta(hours=24))
        start, end = get_start_end(start, end)
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.ResultData.RESULT_STATUS,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            filter_query=filter_query,
            timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
            agg_funcs=["max"],
            group_by=mm_schemas.EventFieldType.ENDPOINT_ID,
            preform_agg_columns=[mm_schemas.ResultData.RESULT_STATUS],
        )
        df.rename(
            columns={
                f"max({mm_schemas.ResultData.RESULT_STATUS})": mm_schemas.ResultData.RESULT_STATUS
            },
            inplace=True,
        )
        if not df.empty:
            df.dropna(inplace=True)
        return df

    def count_results_by_status(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        endpoint_ids: Optional[Union[str, list[str]]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
        result_status_list: Optional[list[int]] = None,
    ) -> dict[tuple[str, int], int]:
        filter_query = ""

        start, end = get_start_end(start=start, end=end, delta=timedelta(hours=24))

        if endpoint_ids:
            filter_query = self._generate_filter_query(
                filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
                filter_values=endpoint_ids,
            )
        if application_names:
            app_filter_query = self._generate_filter_query(
                filter_column=mm_schemas.ApplicationEvent.APPLICATION_NAME,
                filter_values=application_names,
            )
            if filter_query:
                filter_query += f" AND {app_filter_query}"
            else:
                filter_query = app_filter_query
        if result_status_list:
            status_filter_query = self._generate_filter_query(
                filter_column=mm_schemas.ResultData.RESULT_STATUS,
                filter_values=result_status_list,
            )
            if filter_query:
                filter_query += f" AND {status_filter_query}"
            else:
                filter_query = status_filter_query

        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.ResultData.RESULT_STATUS,
                mm_schemas.ResultData.RESULT_VALUE,
            ],
            filter_query=filter_query,
            timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
            group_by=[
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.ResultData.RESULT_STATUS,
            ],
            agg_funcs=["count"],
            preform_agg_columns=[mm_schemas.ResultData.RESULT_VALUE],
        )
        if df.empty:
            return {}

        # Convert DataFrame to a dictionary
        return {
            (
                row[mm_schemas.WriterEvent.APPLICATION_NAME],
                row[mm_schemas.ResultData.RESULT_STATUS],
            ): row["count(result_value)"]
            for _, row in df.iterrows()
        }

    def count_processed_model_endpoints(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
    ) -> dict:
        filter_query = ""
        start, end = get_start_end(start=start, end=end, delta=timedelta(hours=24))

        if application_names:
            filter_query = self._generate_filter_query(
                filter_column=mm_schemas.WriterEvent.APPLICATION_NAME,
                filter_values=application_names,
            )

        def get_application_endpoints_records(super_table: str) -> pd.DataFrame:
            return self._get_records(
                table=super_table,
                start=start,
                end=end,
                timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
                columns=[
                    mm_schemas.WriterEvent.APPLICATION_NAME,
                    mm_schemas.EventFieldType.ENDPOINT_ID,
                ],
                filter_query=filter_query,
                group_by=[
                    mm_schemas.WriterEvent.APPLICATION_NAME,
                    mm_schemas.EventFieldType.ENDPOINT_ID,
                ],
                preform_agg_columns=[mm_schemas.ResultData.RESULT_VALUE],
                agg_funcs=["last"],
            )

        df_results = get_application_endpoints_records(
            super_table=self.tables[
                mm_schemas.TDEngineSuperTables.APP_RESULTS
            ].super_table
        )
        df_metrics = get_application_endpoints_records(
            super_table=self.tables[mm_schemas.TDEngineSuperTables.METRICS].super_table
        )

        combined_df = pd.concat([df_results, df_metrics]).drop_duplicates()

        if combined_df.empty:
            return {}
        grouped_df = combined_df.groupby(
            mm_schemas.WriterEvent.APPLICATION_NAME
        ).count()

        # Convert DataFrame to a dictionary
        return grouped_df[mm_schemas.WriterEvent.ENDPOINT_ID].to_dict()

    def calculate_latest_metrics(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
    ) -> list[
        Union[mm_schemas.ApplicationResultRecord, mm_schemas.ApplicationMetricRecord]
    ]:
        metric_list = []
        filter_query = ""
        start, end = get_start_end(start=start, end=end, delta=timedelta(hours=24))

        if application_names:
            filter_query = self._generate_filter_query(
                filter_column=mm_schemas.WriterEvent.APPLICATION_NAME,
                filter_values=application_names,
            )

        def get_latest_metrics_records(
            record_type: Literal["metrics", "results"],
        ) -> pd.DataFrame:
            columns = [
                mm_schemas.WriterEvent.END_INFER_TIME,
                mm_schemas.WriterEvent.APPLICATION_NAME,
            ]
            if record_type == "results":
                table = self.tables[
                    mm_schemas.TDEngineSuperTables.APP_RESULTS
                ].super_table
                columns += [
                    mm_schemas.ResultData.RESULT_NAME,
                    mm_schemas.ResultData.RESULT_VALUE,
                    mm_schemas.ResultData.RESULT_STATUS,
                    mm_schemas.ResultData.RESULT_KIND,
                ]
                agg_column = mm_schemas.ResultData.RESULT_VALUE
            else:
                table = self.tables[mm_schemas.TDEngineSuperTables.METRICS].super_table
                columns += [
                    mm_schemas.MetricData.METRIC_NAME,
                    mm_schemas.MetricData.METRIC_VALUE,
                ]
                agg_column = mm_schemas.MetricData.METRIC_VALUE

            return self._get_records(
                table=table,
                start=start,
                end=end,
                columns=columns,
                filter_query=filter_query,
                timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
                # Aggregate per application/metric pair regardless of timestamp
                group_by=columns[1:],
                preform_agg_columns=[agg_column],
                agg_funcs=["last"],
            )

        df_results = get_latest_metrics_records(record_type="results")
        df_metrics = get_latest_metrics_records(record_type="metrics")

        if df_results.empty and df_metrics.empty:
            return metric_list

        def build_metric_objects() -> (
            list[
                Union[
                    mm_schemas.ApplicationResultRecord,
                    mm_schemas.ApplicationMetricRecord,
                ]
            ]
        ):
            metric_objects = []
            if not df_results.empty:
                df_results.rename(
                    columns={
                        f"last({mm_schemas.ResultData.RESULT_VALUE})": mm_schemas.ResultData.RESULT_VALUE,
                    },
                    inplace=True,
                )
                for _, row in df_results.iterrows():
                    metric_objects.append(
                        mm_schemas.ApplicationResultRecord(
                            time=datetime.fromisoformat(
                                row[mm_schemas.WriterEvent.END_INFER_TIME].replace(
                                    " +", "+"
                                )
                            ),
                            result_name=row[mm_schemas.ResultData.RESULT_NAME],
                            kind=row[mm_schemas.ResultData.RESULT_KIND],
                            status=row[mm_schemas.ResultData.RESULT_STATUS],
                            value=row[mm_schemas.ResultData.RESULT_VALUE],
                        )
                    )

            if not df_metrics.empty:
                df_metrics.rename(
                    columns={
                        f"last({mm_schemas.MetricData.METRIC_VALUE})": mm_schemas.MetricData.METRIC_VALUE,
                    },
                    inplace=True,
                )
                for _, row in df_metrics.iterrows():
                    metric_objects.append(
                        mm_schemas.ApplicationMetricRecord(
                            time=datetime.fromisoformat(
                                row[mm_schemas.WriterEvent.END_INFER_TIME].replace(
                                    " +", "+"
                                )
                            ),
                            metric_name=row[mm_schemas.MetricData.METRIC_NAME],
                            value=row[mm_schemas.MetricData.METRIC_VALUE],
                        )
                    )

            return metric_objects

        return build_metric_objects()

    def get_metrics_metadata(
        self,
        endpoint_id: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> pd.DataFrame:
        start, end = get_start_end(start, end)
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.METRICS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.ApplicationEvent.APPLICATION_NAME,
                mm_schemas.MetricData.METRIC_NAME,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            filter_query=self._generate_filter_query(
                filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
                filter_values=endpoint_id,
            ),
            timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
            group_by=[
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.MetricData.METRIC_NAME,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            agg_funcs=["last"],
        )
        df.rename(
            columns={
                f"last({mm_schemas.ApplicationEvent.APPLICATION_NAME})": mm_schemas.ApplicationEvent.APPLICATION_NAME,
                f"last({mm_schemas.MetricData.METRIC_NAME})": mm_schemas.MetricData.METRIC_NAME,
                f"last({mm_schemas.EventFieldType.ENDPOINT_ID})": mm_schemas.EventFieldType.ENDPOINT_ID,
            },
            inplace=True,
        )
        if not df.empty:
            df.dropna(inplace=True)
        return df

    def get_results_metadata(
        self,
        endpoint_id: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> pd.DataFrame:
        start, end = get_start_end(start, end)
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.ApplicationEvent.APPLICATION_NAME,
                mm_schemas.ResultData.RESULT_NAME,
                mm_schemas.ResultData.RESULT_KIND,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            filter_query=self._generate_filter_query(
                filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
                filter_values=endpoint_id,
            ),
            timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
            group_by=[
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.ResultData.RESULT_NAME,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            agg_funcs=["last"],
        )
        df.rename(
            columns={
                f"last({mm_schemas.ApplicationEvent.APPLICATION_NAME})": mm_schemas.ApplicationEvent.APPLICATION_NAME,
                f"last({mm_schemas.ResultData.RESULT_NAME})": mm_schemas.ResultData.RESULT_NAME,
                f"last({mm_schemas.ResultData.RESULT_KIND})": mm_schemas.ResultData.RESULT_KIND,
                f"last({mm_schemas.EventFieldType.ENDPOINT_ID})": mm_schemas.EventFieldType.ENDPOINT_ID,
            },
            inplace=True,
        )
        if not df.empty:
            df.dropna(inplace=True)
        return df

    def get_error_count(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> pd.DataFrame:
        filter_query = self._generate_filter_query(
            filter_column=mm_schemas.EventFieldType.ENDPOINT_ID,
            filter_values=endpoint_ids,
        )
        filter_query += f"AND {mm_schemas.EventFieldType.ERROR_TYPE} = '{mm_schemas.EventFieldType.INFER_ERROR}'"
        start, end = get_start_end(start, end)
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.ERRORS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.EventFieldType.MODEL_ERROR,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            agg_funcs=["count"],
            filter_query=filter_query,
            group_by=mm_schemas.EventFieldType.ENDPOINT_ID,
            preform_agg_columns=[mm_schemas.EventFieldType.MODEL_ERROR],
        )
        df.rename(
            columns={f"count({mm_schemas.EventFieldType.MODEL_ERROR})": "error_count"},
            inplace=True,
        )
        if not df.empty:
            df.dropna(inplace=True)
        return df

    def get_avg_latency(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> pd.DataFrame:
        endpoint_ids = (
            endpoint_ids if isinstance(endpoint_ids, list) else [endpoint_ids]
        )
        start, end = get_start_end(start, end, delta=timedelta(hours=24))
        df = self._get_records(
            table=self.tables[mm_schemas.TDEngineSuperTables.PREDICTIONS].super_table,
            start=start,
            end=end,
            columns=[
                mm_schemas.EventFieldType.LATENCY,
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            agg_funcs=["avg"],
            filter_query=f"endpoint_id IN({str(endpoint_ids)[1:-1]})",
            group_by=mm_schemas.EventFieldType.ENDPOINT_ID,
            preform_agg_columns=[mm_schemas.EventFieldType.LATENCY],
        )
        df.rename(
            columns={f"avg({mm_schemas.EventFieldType.LATENCY})": "avg_latency"},
            inplace=True,
        )
        if not df.empty:
            df.dropna(inplace=True)
        return df

    def add_basic_metrics(
        self,
        model_endpoint_objects: list[mlrun.common.schemas.ModelEndpoint],
        metric_list: Optional[list[str]] = None,
    ) -> list[mlrun.common.schemas.ModelEndpoint]:
        """
        Add basic metrics to the model endpoint object.

        :param model_endpoint_objects: A list of `ModelEndpoint` objects that will
                                        be filled with the relevant basic metrics.
        :param metric_list:            List of metrics to include from the time series DB. Defaults to all metrics.

        :return: A list of `ModelEndpointMonitoringMetric` objects.
        """

        uids = [mep.metadata.uid for mep in model_endpoint_objects]

        metric_name_to_function = {
            "error_count": self.get_error_count,
            "last_request": self.get_last_request,
            "avg_latency": self.get_avg_latency,
            "result_status": self.get_drift_status,
        }
        if metric_list is not None:
            for metric_name in list(metric_name_to_function):
                if metric_name not in metric_list:
                    del metric_name_to_function[metric_name]

        metric_name_to_df = {
            metric_name: function(endpoint_ids=uids)
            for metric_name, function in metric_name_to_function.items()
        }

        def add_metrics(
            mep: mlrun.common.schemas.ModelEndpoint,
            df_dictionary: dict[str, pd.DataFrame],
        ):
            for metric in df_dictionary.keys():
                df = df_dictionary.get(metric, pd.DataFrame())
                if not df.empty:
                    line = df[df["endpoint_id"] == mep.metadata.uid]
                    if not line.empty and metric in line:
                        value = line[metric].item()
                        if isinstance(value, pd.Timestamp):
                            value = value.to_pydatetime()
                        setattr(mep.status, metric, value)

            return mep

        return list(
            map(
                lambda mep: add_metrics(
                    mep=mep,
                    df_dictionary=metric_name_to_df,
                ),
                model_endpoint_objects,
            )
        )

    def get_drift_data(
        self,
        start: datetime,
        end: datetime,
    ) -> mm_schemas.ModelEndpointDriftValues:
        filter_query = self._generate_filter_query(
            filter_column=mm_schemas.ResultData.RESULT_STATUS,
            filter_values=[
                mm_schemas.ResultStatusApp.potential_detection.value,
                mm_schemas.ResultStatusApp.detected.value,
            ],
        )
        table = self.tables[mm_schemas.TDEngineSuperTables.APP_RESULTS].super_table
        start, end, interval = self._prepare_aligned_start_end(start, end)

        # get per time-interval x endpoint_id combination the max result status
        df = self._get_records(
            table=table,
            start=start,
            end=end,
            interval=interval,
            columns=[mm_schemas.ResultData.RESULT_STATUS],
            filter_query=filter_query,
            timestamp_column=mm_schemas.WriterEvent.END_INFER_TIME,
            agg_funcs=["max"],
            partition_by=mm_schemas.WriterEvent.ENDPOINT_ID,
        )
        if df.empty:
            return mm_schemas.ModelEndpointDriftValues(values=[])

        df["_wstart"] = pd.to_datetime(df["_wstart"])
        return self._df_to_drift_data(df)

    # Note: this function serves as a reference for checking the TSDB for the existence of a metric.
    #
    # def read_prediction_metric_for_endpoint_if_exists(
    #     self, endpoint_id: str
    # ) -> Optional[mm_schemas.ModelEndpointMonitoringMetric]:
    #     """
    #     Read the "invocations" metric for the provided model endpoint, and return the metric object
    #     if it exists.
    #
    #     :param endpoint_id: The model endpoint identifier.
    #     :return:            `None` if the invocations metric does not exist, otherwise return the
    #                         corresponding metric object.
    #     """
    #     # Read just one record, because we just want to check if there is any data for this endpoint_id
    #     predictions = self.read_predictions(
    #         endpoint_id=endpoint_id,
    #         start=datetime.min,
    #         end=mlrun.utils.now_date(),
    #         limit=1,
    #     )
    #     if predictions:
    #         return get_invocations_metric(self.project)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import zipfile
from typing import Optional, Union

import numpy as np
import tensorflow as tf
from packaging import version
from tensorflow import keras

import mlrun
from mlrun.artifacts import Artifact
from mlrun.features import Feature

from .._common import without_mlrun_interface
from .._dl_common import DLModelHandler
from .mlrun_interface import TFKerasMLRunInterface
from .utils import TFKerasUtils, is_keras_3


class TFKerasModelHandler(DLModelHandler):
    """
    Class for handling a tensorflow.keras model, enabling loading and saving it during runs.
    """

    # Framework name:
    FRAMEWORK_NAME = "tensorflow.keras"

    # Declare a type of input sample (only from keras v3 there is a KerasTensor type):
    IOSample = Union[tf.Tensor, tf.TensorSpec, "keras.KerasTensor", np.ndarray]

    class ModelFormats:
        """
        Model formats to pass to the 'TFKerasModelHandler' for loading and saving keras models.
        """

        SAVED_MODEL = "SavedModel"
        KERAS = "keras"
        H5 = "h5"
        JSON_ARCHITECTURE_H5_WEIGHTS = "json_h5"

        @classmethod
        def default(cls) -> str:
            """
            Get the default model format to use for saving and loading the model based on the keras version.

            :return: The default model format to use.
            """
            return cls.KERAS if is_keras_3() else cls.SAVED_MODEL

    class _LabelKeys:
        """
        Required labels keys to log with the model.
        """

        MODEL_FORMAT = "model-format"
        SAVE_TRACES = "save-traces"

    def __init__(
        self,
        model: keras.Model = None,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        model_format: Optional[str] = None,
        context: mlrun.MLClientCtx = None,
        modules_map: Optional[
            Union[dict[str, Union[None, str, list[str]]], str]
        ] = None,
        custom_objects_map: Optional[
            Union[dict[str, Union[str, list[str]]], str]
        ] = None,
        custom_objects_directory: Optional[str] = None,
        save_traces: bool = False,
        **kwargs,
    ):
        """
        Initialize the handler. The model can be set here so it won't require loading. Notice that if the model path
        given is of a previously logged model (store model object path), all of the other configurations will be loaded
        automatically as they were logged with the model, hence they are optional.

        :param model:                    Model to handle or None in case a loading parameters were supplied.
        :param model_path:               Path to the model's directory to load it from. The model files must start with
                                         the given model name and the directory must contain based on the given model
                                         formats:
                                         * SavedModel - A zip file 'model_name.zip' or a directory named 'model_name'.
                                         * H5 - A h5 file 'model_name.h5'.
                                         * Architecture and weights - The json file 'model_name.json' and h5 weight file
                                           'model_name.h5'.
                                         The model path can be also passed as a model object path in the following
                                         format: 'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'.
        :param model_name:               The model name for saving and logging the model:
                                         * Mandatory for loading the model from a local path.
                                         * If given a logged model (store model path) it will be read from the artifact.
                                         * If given a loaded model object and the model name is None, the name will be
                                           set to the model's object name / class.
        :param model_format:             The format to use for saving and loading the model. Should be passed as a
                                         member of the class 'ModelFormats'.
        :param context:                  MLRun context to work with for logging the model.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key
                                         is a path to a module and its value is the object name to import from it. All
                                         the modules will be imported globally. If multiple objects needed to be
                                         imported from the same module a list can be given. The map can be passed as a
                                         path to a json file as well. For example:
                                         {
                                             "module1": None,  # => import module1
                                             "module2": ["func1", "func2"],  # => from module2 import func1, func2
                                             "module3.sub_module": "func3",  # => from module3.sub_module import func3
                                         }
                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:
                                         {
                                             "/.../custom_optimizer.py": "optimizer",
                                             "/.../custom_layers.py": ["layer1", "layer2"]
                                         }
                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param save_traces:              Whether to use functions saving (only available for the 'SavedModel' format)
                                         for loading the model later without the custom objects dictionary. Only from
                                         tensorflow version >= 2.4.0. Using this setting will increase the model saving
                                         size.

        :raise MLRunInvalidArgumentError: In case the input was incorrect:
                                          * Model format is unrecognized.
                                          * There was no model or model directory supplied.
                                          * 'save_traces' parameter was miss-used.
        """
        # Validate given format:
        if not model_format:
            model_format = TFKerasModelHandler.ModelFormats.default()
        if model_format not in [
            TFKerasModelHandler.ModelFormats.SAVED_MODEL,
            TFKerasModelHandler.ModelFormats.KERAS,
            TFKerasModelHandler.ModelFormats.H5,
            TFKerasModelHandler.ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS,
        ]:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unrecognized model format: '{model_format}'. Please use one of the class members of "
                "'TFKerasModelHandler.ModelFormats'"
            )
        if not is_keras_3():
            if model_format == TFKerasModelHandler.ModelFormats.KERAS:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "The 'keras' model format is only supported in Keras 3.0.0 and above. "
                    f"Current version is {keras.__version__}."
                )
        else:
            if (
                model_format == TFKerasModelHandler.ModelFormats.SAVED_MODEL
                or model_format
                == TFKerasModelHandler.ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The '{model_format}' model format is not supported in Keras 3.0.0 and above. "
                    f"Current version is {keras.__version__}."
                )

        # Validate 'save_traces':
        if save_traces:
            if version.parse(tf.__version__) < version.parse("2.4.0"):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'save_traces' parameter can be true only for tensorflow versions >= 2.4. Current "
                    f"version is {tf.__version__}"
                )
            if model_format != TFKerasModelHandler.ModelFormats.SAVED_MODEL:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "The 'save_traces' parameter is valid only for the 'SavedModel' format."
                )

        # If the model is given without a model name, set the model name:
        if model_name is None and model is not None:
            model_name = model.name

        # Store the configuration:
        self._model_format = model_format
        self._save_traces = save_traces

        # If the model format is architecture and weights, this will hold the weights file collected:
        self._weights_file = None  # type: str

        # Setup the base handler class:
        super().__init__(
            model=model,
            model_path=model_path,
            model_name=model_name,
            modules_map=modules_map,
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
            context=context,
            **kwargs,
        )

        # Set the required labels:
        self.set_labels()

    def set_labels(
        self,
        to_add: Optional[dict[str, Union[str, int, float]]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the labels dictionary of this model artifact. There are required labels that cannot be edited or removed.

        :param to_add:    The labels to add.
        :param to_remove: A list of labels keys to remove.
        """
        # Update the user's labels:
        super().set_labels(to_add=to_add, to_remove=to_remove)

        # Set the required labels:
        self._labels[self._LabelKeys.MODEL_FORMAT] = self._model_format
        if self._model_format == self.ModelFormats.SAVED_MODEL:
            self._labels[self._LabelKeys.SAVE_TRACES] = self._save_traces

    # TODO: output_path won't work well with logging artifacts. Need to look into changing the logic of 'log_artifact'.
    @without_mlrun_interface(interface=TFKerasMLRunInterface)
    def save(
        self, output_path: Optional[str] = None, **kwargs
    ) -> Union[dict[str, Artifact], None]:
        """
        Save the handled model at the given output path. If a MLRun context is available, the saved model files will be
        logged and returned as artifacts.

        :param output_path: The full path to the directory to save the handled model at. If not given, the context
                            stored will be used to save the model in the default artifacts location.

        :return The saved model additional artifacts (if needed) dictionary if context is available and None otherwise.
        """
        super().save(output_path=output_path)

        # Setup the returning model artifacts list:
        artifacts = {}  # type: Dict[str, Artifact]

        # Set the output path:
        if output_path is None:
            output_path = os.path.join(self._context.artifact_path, self._model_name)

        # ModelFormats.H5 - Save as a h5 file:
        if self._model_format == TFKerasModelHandler.ModelFormats.H5:
            self._model_file = f"{self._model_name}.h5"
            self._model.save(self._model_file)

        # ModelFormats.keras - Save as a keras file:
        elif self._model_format == self.ModelFormats.KERAS:
            self._model_file = f"{self._model_name}.keras"
            self._model.save(self._model_file)

        # ModelFormats.SAVED_MODEL - Save as a SavedModel directory and zip its file:
        elif self._model_format == TFKerasModelHandler.ModelFormats.SAVED_MODEL:
            # Save it in a SavedModel format directory:
            # Note: Using keras>=3.0.0 can save in this format via `model.export` but then it won't be able to load it
            # back, only for inference. So, we use the `save` method instead for keras 2 and validate the user won't use
            # keras 3 and this model format.
            if self._save_traces is True:
                # Save traces can only be used in versions >= 2.4, so only if it's true, we use it in the call:
                self._model.save(self._model_name, save_traces=self._save_traces)
            else:
                self._model.save(self._model_name)
            # Zip it:
            self._model_file = f"{self._model_name}.zip"
            shutil.make_archive(
                base_name=self._model_name, format="zip", base_dir=self._model_name
            )

        # ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS - Save as a json architecture and h5 weights files:
        else:
            # Save the model architecture (json):
            model_architecture = self._model.to_json()
            self._model_file = f"{self._model_name}.json"
            with open(self._model_file, "w") as json_file:
                json_file.write(model_architecture)
            # Save the model weights (h5):
            self._weights_file = f"{self._model_name}.h5"
            self._model.save_weights(self._weights_file)

        # Update the paths and log artifacts if context is available:
        if self._weights_file is not None:
            if self._context is not None:
                artifacts[self._get_weights_file_artifact_name()] = (
                    self._context.log_artifact(
                        self._weights_file,
                        local_path=self._weights_file,
                        artifact_path=output_path,
                        db_key=False,
                    )
                )

        return artifacts if self._context is not None else None

    def load(self, checkpoint: Optional[str] = None, **kwargs):
        """
        Load the specified model in this handler. If a checkpoint is required to be loaded, it can be given here
        according to the provided model path in the initialization of this handler. Additional parameters for the class
        initializer can be passed via the args list and kwargs dictionary.

        :param checkpoint: The checkpoint label to load the weights from. If the model path is of a store object, the
                           checkpoint will be taken from the logged checkpoints artifacts logged with the model. If the
                           model path is of a local directory, the checkpoint will be searched in it by the provided
                           name to this parameter.
        """
        # TODO: Add support for checkpoint loading after creating MLRun's checkpoint callback.
        if checkpoint is not None:
            raise NotImplementedError(
                "Loading a model using checkpoint is not yet implemented."
            )

        super().load()

        # ModelFormats.H5 - Load from a h5 file:
        if self._model_format == TFKerasModelHandler.ModelFormats.H5:
            self._model = keras.models.load_model(
                self._model_file, custom_objects=self._custom_objects
            )

        # ModelFormats.KERAS - Load from a keras file:
        elif self._model_format == TFKerasModelHandler.ModelFormats.KERAS:
            self._model = keras.models.load_model(
                self._model_file, custom_objects=self._custom_objects
            )

        # ModelFormats.SAVED_MODEL - Load from a SavedModel directory:
        elif self._model_format == TFKerasModelHandler.ModelFormats.SAVED_MODEL:
            self._model = keras.models.load_model(
                self._model_file, custom_objects=self._custom_objects
            )

        # ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS - Load from a json architecture file and a h5 weights file:
        else:
            # Load the model architecture (json):
            with open(self._model_file) as json_file:
                model_architecture = json_file.read()
            self._model = keras.models.model_from_json(
                model_architecture, custom_objects=self._custom_objects
            )
            # Load the model weights (h5):
            self._model.load_weights(self._weights_file)

    def to_onnx(
        self,
        model_name: Optional[str] = None,
        optimize: bool = True,
        input_signature: Union[
            list[tf.TensorSpec], list[np.ndarray], tf.TensorSpec, np.ndarray
        ] = None,
        output_path: Optional[str] = None,
        log: Optional[bool] = None,
    ):
        """
        Convert the model in this handler to an ONNX model.
        :param model_name:      The name to give to the converted ONNX model. If not given the default name will be the
                                stored model name with the suffix '_onnx'.
        :param optimize:        Whether to optimize the ONNX model using 'onnxoptimizer' before saving the model.
                                Default: True.
        :param input_signature: An numpy.ndarray or tensorflow.TensorSpec that describe the input port (shape and data
                                type). If the model has multiple inputs, a list is expected in the order of the input
                                ports. If not provided, the method will try to extract the input signature of the model.
        :param output_path:     In order to save the ONNX model, pass here the output directory. The model file will be
                                named with the model name given. Default: None (not saving).
        :param log:             In order to log the ONNX model, pass True. If None, the model will be logged if this
                                handler has a MLRun context set. Default: None.

        :return: The converted ONNX model (onnx.ModelProto).

        :raise MLRunMissingDependencyError: If the onnx modules are missing in the interpreter.
        :raise MLRunInvalidArgumentError:   If the input signatures was not given and inputs are not set in the handler.
        """
        # Import onnx related modules:
        try:
            import tf2onnx

            from mlrun.frameworks.onnx import ONNXModelHandler
        except ModuleNotFoundError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "ONNX conversion requires additional packages to be installed. "
                "Please run 'pip install mlrun[tensorflow]' to install MLRun's Tensorflow package."
            )

        # Set the onnx model name:
        model_name = self._get_default_onnx_model_name(model_name=model_name)

        # Set the input signature:
        if input_signature is None:
            if self._inputs is not None:
                # Parse the set input features:
                input_signature = [
                    tf.TensorSpec(
                        shape=input_feature.dims,
                        dtype=TFKerasUtils.convert_value_type_to_tf_dtype(
                            value_type=input_feature.value_type
                        ),
                    )
                    for input_feature in self._inputs
                ]
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "In order to convert the model to ONNX the 'input_signature' must be given or you can use one of "
                    "the 'set_inputs', 'read_inputs_from_model' methods."
                )
        elif not isinstance(input_signature, list):
            # Wrap it in a list:
            input_signature = [input_signature]

        # Set the output path:
        if output_path is not None:
            output_path = os.path.join(output_path, f"{model_name}.onnx")

        # Set the logging flag:
        log = self._context is not None if log is None else log

        # Convert to ONNX:
        model_proto, external_tensor_storage = tf2onnx.convert.from_keras(
            model=self._model, input_signature=input_signature, output_path=output_path
        )

        # Create a handler for the ONNX model:
        onnx_handler = ONNXModelHandler(
            model_name=model_name, model=model_proto, context=self._context
        )

        # Pass on the inputs and outputs properties:
        if self._inputs is not None:
            onnx_handler.set_inputs(features=self._inputs)
        if self._outputs is not None:
            onnx_handler.set_outputs(features=self._outputs)

        # Optimize the model if needed:
        if optimize:
            onnx_handler.optimize()
            # Save if logging is not required, as logging will save as well:
            if not log and output_path is not None:
                onnx_handler.save(output_path=output_path)

        # Log as a model object if needed:
        if log:
            onnx_handler.log()

        return onnx_handler.model

    def read_inputs_from_model(self):
        """
        Extract the inputs information out of the model and set it in the handler.

        :raise MLRunRuntimeError: If there is no model in this handler.
        """
        # Validate there is a model available:
        if self._model is None:
            raise mlrun.errors.MLRunRuntimeError(
                "The model in this handler was not loaded or given in initialization so the inputs cannot be read."
            )

        # Read the inputs:
        input_signature = [
            getattr(input_layer, "type_spec", input_layer)
            for input_layer in self._model.inputs
        ]

        # Set the inputs:
        self.set_inputs(from_sample=input_signature)

    def read_outputs_from_model(self):
        """
        Extract the outputs information out of the model and set it in the handler.

        :raise MLRunRuntimeError: If there is no model in this handler.
        """
        # Validate there is a model available:
        if self._model is None:
            raise mlrun.errors.MLRunRuntimeError(
                "The model in this handler was not loaded or given in initialization so the outputs cannot be read."
            )

        # Read the outputs:
        output_signature = [
            getattr(output_layer, "type_spec", output_layer)
            for output_layer in self._model.outputs
        ]

        # Set the outputs:
        self.set_outputs(from_sample=output_signature)

    def _collect_files_from_store_object(self):
        """
        If the model path given is of a store object, collect the needed model files into this handler for later loading
        the model.
        """
        # Read the settings:
        self._model_format = self._model_artifact.labels[self._LabelKeys.MODEL_FORMAT]
        self._save_traces = self._model_artifact.labels.get(
            self._LabelKeys.SAVE_TRACES, None
        )

        # Read additional files according to the model format used:
        if self._model_format == TFKerasModelHandler.ModelFormats.SAVED_MODEL:
            # Unzip the SavedModel directory:
            with zipfile.ZipFile(self._model_file, "r") as zip_file:
                zip_file.extractall(os.path.dirname(self._model_file))
            # Set the model file to the unzipped directory:
            self._model_file = os.path.join(
                os.path.dirname(self._model_file), self._model_name
            )
        elif self._model_format == TFKerasModelHandler.ModelFormats.KERAS:
            # Rename the model file suffix:
            self._rename_model_file_suffix(suffix="keras")
        elif self._model_format == TFKerasModelHandler.ModelFormats.H5:
            # Rename the model file suffix:
            self._rename_model_file_suffix(suffix="h5")
        elif (  # ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS
            self._model_format
            == TFKerasModelHandler.ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS
        ):
            # Rename the model file suffix:
            self._rename_model_file_suffix(suffix="json")
            # Get the weights file:
            self._weights_file = self._extra_data[
                self._get_weights_file_artifact_name()
            ].local()

        # Continue collecting from abstract class:
        super()._collect_files_from_store_object()

    def _rename_model_file_suffix(self, suffix: str):
        """
        Rename the model file suffix to the given one.

        This is used for the case of loading a model from a store object that was saved with a different suffix as when
        keras tries to load it, it validates the suffix. The `artifacts.model.get_model` function is downloading the
        file to a temp file with a `pkl` suffix, so it needs to be replaced:than the one keras expects.

        :param suffix: The suffix to rename the model file to (without the trailing dot).
        """
        new_name = self._model_file.rsplit(".", 1)[0] + f".{suffix}"
        os.rename(self._model_file, new_name)
        self._model_file = new_name

    def _collect_files_from_local_path(self):
        """
        If the model path given is of a local path, search for the needed model files and collect them into this handler
        for later loading the model.

        :raise MLRunNotFoundError: If any of the required files are missing.
        """
        # ModelFormats.H5 - Get the h5 model file:
        if self._model_format == TFKerasModelHandler.ModelFormats.H5:
            self._model_file = os.path.join(self._model_path, f"{self._model_name}.h5")
            if not os.path.exists(self._model_file):
                raise mlrun.errors.MLRunNotFoundError(
                    f"The model file '{self._model_name}.h5' was not found within the given 'model_path': "
                    f"'{self._model_path}'"
                )

        # ModelFormats.KERAS - Get the keras model file:
        elif self._model_format == TFKerasModelHandler.ModelFormats.KERAS:
            self._model_file = os.path.join(
                self._model_path, f"{self._model_name}.keras"
            )
            if not os.path.exists(self._model_file):
                raise mlrun.errors.MLRunNotFoundError(
                    f"The model file '{self._model_name}.keras' was not found within the given 'model_path': "
                    f"'{self._model_path}'"
                )

        # ModelFormats.SAVED_MODEL - Get the zip file and extract it, or simply locate the directory:
        elif self._model_format == TFKerasModelHandler.ModelFormats.SAVED_MODEL:
            self._model_file = os.path.join(self._model_path, f"{self._model_name}.zip")
            if os.path.exists(self._model_file):
                # Unzip it:
                with zipfile.ZipFile(self._model_file, "r") as zip_file:
                    zip_file.extractall(os.path.dirname(self._model_file))
                # Set the model file to the unzipped directory:
                self._model_file = os.path.join(
                    os.path.dirname(self._model_file), self._model_name
                )
            else:
                # Look for the SavedModel directory:
                self._model_file = os.path.join(self._model_path, self._model_name)
                if not os.path.exists(self._model_file):
                    raise mlrun.errors.MLRunNotFoundError(
                        f"There is no SavedModel zip archive '{self._model_name}.zip' or a SavedModel directory named "
                        f"'{self._model_name}' the given 'model_path': '{self._model_path}'"
                    )

        # ModelFormats.JSON_ARCHITECTURE_H5_WEIGHTS - Save as a json architecture and h5 weights files:
        else:
            # Locate the model architecture json file:
            self._model_file = f"{self._model_name}.json"
            if not os.path.exists(os.path.join(self._model_path, self._model_file)):
                raise mlrun.errors.MLRunNotFoundError(
                    f"The model architecture file '{self._model_file}' is missing in the given 'model_path': "
                    f"'{self._model_path}'"
                )
            # Locate the model weights h5 file:
            self._weights_file = f"{self._model_name}.h5"
            if not os.path.exists(os.path.join(self._model_path, self._weights_file)):
                raise mlrun.errors.MLRunNotFoundError(
                    f"The model weights file '{self._weights_file}' is missing in the given 'model_path': "
                    f"'{self._model_path}'"
                )

    def _read_sample(self, sample: IOSample) -> Feature:
        """
        Read the sample into a MLRun Feature.

        :param sample: The sample to read.

        :return: The created Feature.

        :raise MLRunInvalidArgumentError: In case the given sample type cannot be read.
        """
        # Supported types:
        if isinstance(sample, np.ndarray):
            return super()._read_sample(sample=sample)
        elif isinstance(sample, tf.TensorSpec) or (
            is_keras_3() and isinstance(sample, keras.KerasTensor)
        ):
            return Feature(
                name=sample.name,
                value_type=TFKerasUtils.convert_tf_dtype_to_value_type(
                    tf_dtype=sample.dtype
                ),
                dims=list(sample.shape),
            )
        elif isinstance(sample, tf.Tensor):
            return Feature(
                value_type=TFKerasUtils.convert_tf_dtype_to_value_type(
                    tf_dtype=sample.dtype
                ),
                dims=list(sample.shape),
            )

        # Unsupported type:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The sample type given '{type(sample)}' is not supported. The input / output ports are readable from "
            f"samples of the following types: tf.Tensor, tf.TensorSpec, np.ndarray."
        )



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.common.types
import mlrun_pipelines.common.ops
import mlrun_pipelines.models

from .base import ObjectFormat


class PipelineFormat(ObjectFormat, mlrun.common.types.StrEnum):
    full = "full"
    metadata_only = "metadata_only"
    name_only = "name_only"
    summary = "summary"

    @staticmethod
    def format_method(_format: str) -> typing.Optional[typing.Callable]:
        def _full(run: mlrun_pipelines.models.PipelineRun) -> dict:
            return run.to_dict()

        def _metadata_only(run: mlrun_pipelines.models.PipelineRun) -> dict:
            return mlrun.utils.helpers.format_run(run, with_project=True)

        def _name_only(run: mlrun_pipelines.models.PipelineRun) -> str:
            return run.get("name")

        def _summary(run: mlrun_pipelines.models.PipelineRun) -> dict:
            return mlrun_pipelines.common.ops.format_summary_from_kfp_run(
                run, run["project"]
            )

        return {
            PipelineFormat.full: _full,
            PipelineFormat.metadata_only: _metadata_only,
            PipelineFormat.name_only: _name_only,
            PipelineFormat.summary: _summary,
        }[_format]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mlrun.common.types

IMAGE_NAME_ENRICH_REGISTRY_PREFIX = "."  # prefix for image name to enrich with registry
MLRUN_SERVING_CONF = "serving-conf"
MLRUN_SERVING_SPEC_MOUNT_PATH = f"/tmp/mlrun/{MLRUN_SERVING_CONF}"
MLRUN_SERVING_SPEC_FILENAME = "serving_spec.json"
MLRUN_SERVING_SPEC_PATH = (
    f"{MLRUN_SERVING_SPEC_MOUNT_PATH}/{MLRUN_SERVING_SPEC_FILENAME}"
)
MLRUN_FUNCTIONS_ANNOTATION = "mlrun/mlrun-functions"
MYSQL_MEDIUMBLOB_SIZE_BYTES = 16 * 1024 * 1024
MLRUN_LABEL_PREFIX = "mlrun/"
DASK_LABEL_PREFIX = "dask.org/"
NUCLIO_LABEL_PREFIX = "nuclio.io/"
RESERVED_TAG_NAME_LATEST = "latest"

# Kubernetes DNS-1123 label name length limit
K8S_DNS_1123_LABEL_MAX_LENGTH = 63


RESERVED_BATCH_JOB_SUFFIX = "-batch"

JOB_TYPE_WORKFLOW_RUNNER = "workflow-runner"
JOB_TYPE_PROJECT_LOADER = "project-loader"
JOB_TYPE_RERUN_WORKFLOW_RUNNER = "rerun-workflow-runner"
MLRUN_ACTIVE_PROJECT = "MLRUN_ACTIVE_PROJECT"


class MLRunInternalLabels:
    ### dask
    dask_cluster_name = f"{DASK_LABEL_PREFIX}cluster-name"
    dask_component = f"{DASK_LABEL_PREFIX}component"

    ### spark
    spark_role = "spark-role"

    ### mpi
    mpi_job_name = "mpi-job-name"
    mpi_job_role = "mpi-job-role"
    mpi_role_type = "mpi_role_type"

    ### nuclio
    nuclio_project_name = f"{NUCLIO_LABEL_PREFIX}project-name"
    nuclio_function_name = f"{NUCLIO_LABEL_PREFIX}function-name"
    nuclio_class = f"{NUCLIO_LABEL_PREFIX}class"

    ### mlrun
    mlrun_auth_key = "mlrun-auth-key"
    mlrun_class = f"{MLRUN_LABEL_PREFIX}class"
    client_python_version = f"{MLRUN_LABEL_PREFIX}client_python_version"
    client_version = f"{MLRUN_LABEL_PREFIX}client_version"
    function = f"{MLRUN_LABEL_PREFIX}function"
    job = f"{MLRUN_LABEL_PREFIX}job"
    name = f"{MLRUN_LABEL_PREFIX}name"
    mlrun_owner = f"{MLRUN_LABEL_PREFIX}owner"
    owner_domain = f"{MLRUN_LABEL_PREFIX}owner_domain"
    project = f"{MLRUN_LABEL_PREFIX}project"
    runner_pod = f"{MLRUN_LABEL_PREFIX}runner-pod"
    schedule_name = f"{MLRUN_LABEL_PREFIX}schedule-name"
    scrape_metrics = f"{MLRUN_LABEL_PREFIX}scrape-metrics"
    tag = f"{MLRUN_LABEL_PREFIX}tag"
    uid = f"{MLRUN_LABEL_PREFIX}uid"
    retry = f"{MLRUN_LABEL_PREFIX}retry-attempt"
    username = f"{MLRUN_LABEL_PREFIX}username"
    username_domain = f"{MLRUN_LABEL_PREFIX}username_domain"
    task_name = f"{MLRUN_LABEL_PREFIX}task-name"
    resource_name = f"{MLRUN_LABEL_PREFIX}resource_name"
    created = f"{MLRUN_LABEL_PREFIX}created"
    producer_type = f"{MLRUN_LABEL_PREFIX}producer-type"
    app_name = f"{MLRUN_LABEL_PREFIX}app-name"
    endpoint_id = f"{MLRUN_LABEL_PREFIX}endpoint-id"
    endpoint_name = f"{MLRUN_LABEL_PREFIX}endpoint-name"
    host = "host"
    job_type = "job-type"
    kind = "kind"
    component = "component"
    mlrun_type = "mlrun__type"
    original_workflow_id = "original-workflow-id"
    workflow_id = "workflow-id"
    retrying = "retrying"
    rerun_counter = "rerun-counter"
    rerun_index = "rerun-index"

    owner = "owner"
    v3io_user = "v3io_user"
    workflow = "workflow"
    feature_vector = "feature-vector"

    @classmethod
    def all(cls):
        return [
            value
            for key, value in cls.__dict__.items()
            if not key.startswith("__") and isinstance(value, str)
        ]

    @staticmethod
    def default_run_labels_to_enrich():
        return [
            MLRunInternalLabels.owner,
            MLRunInternalLabels.v3io_user,
        ]


class DeployStatusTextKind(mlrun.common.types.StrEnum):
    logs = "logs"
    events = "events"


class WorkflowSubmitMode(mlrun.common.types.StrEnum):
    direct = "direct"  # call KFP retry API directly
    rerun = "rerun"  # launch a RerunRunner function



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import concurrent.futures
import datetime
import json
import os
import traceback
import warnings
from collections.abc import Iterator
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any, Final, NamedTuple, Optional, Union, cast

import nuclio_sdk
import numpy as np
import pandas as pd

import mlrun
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.feature_store as fstore
import mlrun.model_monitoring
import mlrun.model_monitoring.db._schedules as schedules
import mlrun.model_monitoring.helpers
import mlrun.platforms.iguazio
from mlrun.common.schemas import EndpointType
from mlrun.common.schemas.model_monitoring.constants import (
    ControllerEvent,
    ControllerEventEndpointPolicy,
)
from mlrun.errors import err_to_str
from mlrun.model_monitoring.helpers import batch_dict2timedelta
from mlrun.utils import datetime_now, logger

_SECONDS_IN_DAY = int(datetime.timedelta(days=1).total_seconds())
_SECONDS_IN_MINUTE = 60


class _Interval(NamedTuple):
    start: datetime.datetime
    end: datetime.datetime


class _BatchWindow:
    TIMESTAMP_RESOLUTION_MICRO: Final = 1e-6  # 0.000001 seconds or 1 microsecond

    def __init__(
        self,
        *,
        schedules_file: schedules.ModelMonitoringSchedulesFileEndpoint,
        application: str,
        timedelta_seconds: int,
        last_updated: float,
        first_request: float,
        endpoint_mode: mm_constants.EndpointMode = mm_constants.EndpointMode.REAL_TIME,
    ) -> None:
        """
        Initialize a batch window object that handles the batch interval time range
        for the monitoring functions.
        All the time values are in seconds.
        The start and stop time are in seconds since the epoch.
        """
        self._application = application
        self._first_request = first_request
        self._stop = last_updated
        self._step = timedelta_seconds
        self._db = schedules_file
        self._endpoint_mode = endpoint_mode
        self._start = self._get_last_analyzed()

    def _get_saved_last_analyzed(
        self,
    ) -> Optional[float]:
        return self._db.get_application_time(self._application)

    def _update_last_analyzed(self, last_analyzed: float) -> None:
        self._db.update_application_time(
            application=self._application, timestamp=last_analyzed
        )

    def _get_initial_last_analyzed(self) -> float:
        if self._endpoint_mode == mm_constants.EndpointMode.BATCH:
            logger.info(
                "No last analyzed time was found for this endpoint and application, as this is "
                "probably the first time this application is running. Initializing last analyzed "
                "to the start of the batch time",
                application=self._application,
                start_batch_time=self._first_request,
            )
            return self._first_request
        logger.info(
            "No last analyzed time was found for this endpoint and application, as this is "
            "probably the first time this application is running. Initializing last analyzed "
            "to the latest between first request the latest between first request time or last "
            "update time minus one day",
            application=self._application,
            first_request=self._first_request,
            last_updated=self._stop,
        )
        # max between one day and the base period
        first_period_in_seconds = max(_SECONDS_IN_DAY, self._step)
        return max(
            self._first_request,
            self._stop - first_period_in_seconds,
        )

    def _get_last_analyzed(self) -> float:
        saved_last_analyzed = self._get_saved_last_analyzed()
        if saved_last_analyzed is not None:
            if self._endpoint_mode == mm_constants.EndpointMode.BATCH:
                # Use the maximum between the saved last analyzed and the start of the batch
                return max(saved_last_analyzed, self._first_request)
            return saved_last_analyzed
        else:
            last_analyzed = self._get_initial_last_analyzed()
            # Update the in-memory DB to avoid duplicate initializations
            self._update_last_analyzed(last_analyzed)
        return last_analyzed

    def get_intervals(self) -> Iterator[_Interval]:
        """Generate the batch interval time ranges."""
        entered = False
        last_analyzed = None
        # Iterate timestamp from start until timestamp <= stop - step
        # so that the last interval will end at (timestamp + step) <= stop.
        # Add 1 to stop - step to get <= and not <.
        for timestamp in np.arange(
            self._start, self._stop - self._step + 1, self._step
        ):
            entered = True
            start_time = datetime.datetime.fromtimestamp(
                timestamp, tz=datetime.timezone.utc
            )
            end_time = datetime.datetime.fromtimestamp(
                timestamp - self.TIMESTAMP_RESOLUTION_MICRO + self._step,
                tz=datetime.timezone.utc,
            )
            yield _Interval(start_time, end_time)

            last_analyzed = timestamp + self._step
            self._update_last_analyzed(last_analyzed)
            logger.debug(
                "Updated the last analyzed time for this endpoint and application",
                application=self._application,
                last_analyzed=last_analyzed,
            )

        if self._endpoint_mode == mm_constants.EndpointMode.BATCH:
            # If the endpoint is a batch endpoint, we need to update the last analyzed time
            # to the end of the batch time.
            if last_analyzed:
                if last_analyzed - self.TIMESTAMP_RESOLUTION_MICRO < self._stop:
                    # If the last analyzed time is earlier than the stop time,
                    # yield the final partial interval from last_analyzed to stop
                    yield _Interval(
                        datetime.datetime.fromtimestamp(
                            last_analyzed, tz=datetime.timezone.utc
                        ),
                        datetime.datetime.fromtimestamp(
                            self._stop, tz=datetime.timezone.utc
                        ),
                    )
            else:
                # The time span between the start and end of the batch is shorter than the step,
                # so we need to yield a partial interval covering that range.
                yield _Interval(
                    datetime.datetime.fromtimestamp(
                        self._start, tz=datetime.timezone.utc
                    ),
                    datetime.datetime.fromtimestamp(
                        self._stop, tz=datetime.timezone.utc
                    ),
                )

            self._update_last_analyzed(last_analyzed=self._stop)
            logger.debug(
                "Updated the last analyzed time for this endpoint and application to the end of the batch time",
                application=self._application,
                last_analyzed=self._stop,
            )

        if not entered:
            logger.debug(
                "All the data is set, but no complete intervals were found. "
                "Wait for last_updated to be updated",
                application=self._application,
                start=self._start,
                stop=self._stop,
                step=self._step,
            )


class _BatchWindowGenerator(AbstractContextManager):
    def __init__(
        self, project: str, endpoint_id: str, window_length: Optional[int] = None
    ) -> None:
        """
        Initialize a batch window generator object that generates batch window objects
        for the monitoring functions.
        """
        self.batch_window: _BatchWindow = None
        self._project = project
        self._endpoint_id = endpoint_id
        self._timedelta = window_length
        self._schedules_file = schedules.ModelMonitoringSchedulesFileEndpoint(
            project=project, endpoint_id=endpoint_id
        )

    def __enter__(self) -> "_BatchWindowGenerator":
        self._schedules_file.__enter__()
        return super().__enter__()

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self._schedules_file.__exit__(
            exc_type=exc_type, exc_value=exc_value, traceback=traceback
        )

    def get_application_list(self) -> set[str]:
        return self._schedules_file.get_application_list()

    def get_min_last_analyzed(self) -> Optional[float]:
        return self._schedules_file.get_min_timestamp()

    @classmethod
    def _get_last_updated_time(
        cls,
        last_request: datetime.datetime,
        endpoint_mode: mm_constants.EndpointMode,
        not_old_batch_endpoint: bool,
    ) -> float:
        """
        Get the last updated time of a model endpoint.
        """

        if endpoint_mode == mm_constants.EndpointMode.REAL_TIME:
            last_updated = last_request.timestamp() - cast(
                float,
                mlrun.mlconf.model_endpoint_monitoring.parquet_batching_timeout_secs,
            )
            if not not_old_batch_endpoint:
                # If the endpoint does not have a stream, `last_updated` should be
                # the minimum between the current time and the last updated time.
                # This compensates for the bumping mechanism - see
                # `update_model_endpoint_last_request`.
                last_updated = min(datetime_now().timestamp(), last_updated)
                logger.debug(
                    "The endpoint does not have a stream", last_updated=last_updated
                )

            return last_updated
        return last_request.timestamp()

    def get_intervals(
        self,
        *,
        application: str,
        first_request: datetime.datetime,
        last_request: datetime.datetime,
        endpoint_mode: mm_constants.EndpointMode,
        not_old_batch_endpoint: bool,
    ) -> Iterator[_Interval]:
        """
        Get the batch window for a specific endpoint and application.
        `first_request` and `last_request` are the timestamps of the first request and last
        request to the endpoint, respectively. They are guaranteed to be nonempty at this point.
        """

        self.batch_window = _BatchWindow(
            schedules_file=self._schedules_file,
            application=application,
            timedelta_seconds=self._timedelta,
            last_updated=self._get_last_updated_time(
                last_request, endpoint_mode, not_old_batch_endpoint
            ),
            first_request=first_request.timestamp(),
            endpoint_mode=endpoint_mode,
        )
        yield from self.batch_window.get_intervals()


def _get_window_length() -> int:
    """Get the timedelta in seconds from the batch dictionary"""
    return int(
        batch_dict2timedelta(
            json.loads(
                cast(str, os.getenv(mm_constants.EventFieldType.BATCH_INTERVALS_DICT))
            )
        ).total_seconds()
    )


class MonitoringApplicationController:
    """
    The main object to handle the monitoring processing job. This object is used to get the required configurations and
    to manage the main monitoring drift detection process based on the current batch.
    Note that the MonitoringApplicationController object requires access keys along with valid project configurations.
    """

    _MAX_FEATURE_SET_PER_WORKER = 1000

    def __init__(self) -> None:
        """Initialize Monitoring Application Controller"""
        self.project = cast(str, mlrun.mlconf.active_project)
        self.project_obj = mlrun.get_run_db().get_project(name=self.project)
        logger.debug(f"Initializing {self.__class__.__name__}", project=self.project)

        self._window_length = _get_window_length()

        self.model_monitoring_access_key = self._get_model_monitoring_access_key()
        self.v3io_access_key = mlrun.mlconf.get_v3io_access_key()
        store, _, _ = mlrun.store_manager.get_or_create_store(
            mlrun.mlconf.artifact_path
        )
        self.storage_options = store.get_storage_options()
        self._controller_stream: Optional[
            Union[
                mlrun.platforms.iguazio.OutputStream,
                mlrun.platforms.iguazio.KafkaOutputStream,
            ]
        ] = None
        self._model_monitoring_stream: Optional[
            Union[
                mlrun.platforms.iguazio.OutputStream,
                mlrun.platforms.iguazio.KafkaOutputStream,
            ]
        ] = None
        self.applications_streams: dict[
            str,
            Union[
                mlrun.platforms.iguazio.OutputStream,
                mlrun.platforms.iguazio.KafkaOutputStream,
            ],
        ] = {}
        self.feature_sets: collections.OrderedDict[
            str, mlrun.feature_store.FeatureSet
        ] = collections.OrderedDict()
        self.tsdb_connector = mlrun.model_monitoring.get_tsdb_connector(
            project=self.project
        )

    @property
    def controller_stream(
        self,
    ) -> Union[
        mlrun.platforms.iguazio.OutputStream,
        mlrun.platforms.iguazio.KafkaOutputStream,
    ]:
        if self._controller_stream is None:
            self._controller_stream = mlrun.model_monitoring.helpers.get_output_stream(
                project=self.project,
                function_name=mm_constants.MonitoringFunctionNames.APPLICATION_CONTROLLER,
                v3io_access_key=self.v3io_access_key,
            )
        return self._controller_stream

    @property
    def model_monitoring_stream(
        self,
    ) -> Union[
        mlrun.platforms.iguazio.OutputStream,
        mlrun.platforms.iguazio.KafkaOutputStream,
    ]:
        if self._model_monitoring_stream is None:
            self._model_monitoring_stream = (
                mlrun.model_monitoring.helpers.get_output_stream(
                    project=self.project,
                    function_name=mm_constants.MonitoringFunctionNames.STREAM,
                    v3io_access_key=self.model_monitoring_access_key,
                )
            )
        return self._model_monitoring_stream

    @staticmethod
    def _get_model_monitoring_access_key() -> Optional[str]:
        access_key = os.getenv(mm_constants.ProjectSecretKeys.ACCESS_KEY)
        # allow access key to be empty and don't fetch v3io access key if not needed
        if access_key is None:
            access_key = mlrun.mlconf.get_v3io_access_key()
        return access_key

    def _should_monitor_endpoint(
        self,
        endpoint: mlrun.common.schemas.ModelEndpoint,
        application_names: set,
        base_period_minutes: int,
        schedules_file: schedules.ModelMonitoringSchedulesFileChief,
    ) -> bool:
        """
        checks if there is a need to monitor the given endpoint, we should monitor endpoint if it stands in the
        next conditions:
            1.  monitoring_mode is enabled
            2.  first request exists
            3.  last request exists
            4.  endpoint_type is not ROUTER
        if the four above conditions apply we require one of the two condition monitor:
            1.  never monitored the one of the endpoint applications meaning min_last_analyzed is None
            2.  min_last_analyzed stands in the condition for sending NOP event and this the first time regular event
            is sent with the combination of  current last_request  & current last_analyzed  per endpoint.
        """
        last_timestamp_sent = schedules_file.get_endpoint_last_request(
            endpoint.metadata.uid
        )
        last_analyzed_sent = schedules_file.get_endpoint_last_analyzed(
            endpoint.metadata.uid
        )
        logger.debug(
            "Chief should monitor endpoint check",
            last_timestamp_sent=last_timestamp_sent,
            last_analyzed_sent=last_analyzed_sent,
            uid=endpoint.metadata.uid,
        )
        if (
            # Is the model endpoint monitored?
            endpoint.status.monitoring_mode == mm_constants.ModelMonitoringMode.enabled
            # Was the model endpoint called? I.e., are the first and last requests nonempty?
            and endpoint.status.first_request
            and endpoint.status.last_request
            # Is the model endpoint not a router endpoint? Router endpoint has no feature stats
            and endpoint.metadata.endpoint_type.value
            != mm_constants.EndpointType.ROUTER.value
        ):
            with _BatchWindowGenerator(
                project=endpoint.metadata.project,
                endpoint_id=endpoint.metadata.uid,
            ) as batch_window_generator:
                current_time = mlrun.utils.datetime_now()
                current_min_last_analyzed = (
                    batch_window_generator.get_min_last_analyzed()
                )
                if (
                    # Different application names, or last analyzed never updated while there are application to monitor
                    application_names
                    and (
                        application_names
                        != batch_window_generator.get_application_list()
                        or not current_min_last_analyzed
                    )
                ):
                    return True
                elif (
                    # Does nop event will be sent to close the relevant window
                    self._should_send_nop_event(
                        base_period_minutes, current_min_last_analyzed, current_time
                    )
                    and (
                        endpoint.status.last_request.timestamp() != last_timestamp_sent
                        or current_min_last_analyzed != last_analyzed_sent
                    )
                ):
                    # Write to schedule chief file the last_request, min_last_analyzed we pushed event to stream
                    schedules_file.update_endpoint_timestamps(
                        endpoint_uid=endpoint.metadata.uid,
                        last_request=endpoint.status.last_request.timestamp(),
                        last_analyzed=current_min_last_analyzed,
                    )
                    return True
                else:
                    logger.info(
                        "All the possible intervals were already analyzed, didn't push regular event",
                        endpoint_id=endpoint.metadata.uid,
                        last_analyzed=current_min_last_analyzed,
                        last_request=endpoint.status.last_request,
                    )
        else:
            logger.info(
                "Should not monitor model endpoint, didn't push regular event",
                endpoint_id=endpoint.metadata.uid,
                endpoint_name=endpoint.metadata.name,
                last_request=endpoint.status.last_request,
                first_request=endpoint.status.first_request,
                endpoint_type=endpoint.metadata.endpoint_type,
                feature_set_uri=endpoint.spec.monitoring_feature_set_uri,
            )
        return False

    @staticmethod
    def _should_send_nop_event(
        base_period_minutes: int,
        min_last_analyzed: float,
        current_time: datetime.datetime,
    ):
        if min_last_analyzed:
            return (
                current_time.timestamp() - min_last_analyzed
                >= datetime.timedelta(minutes=base_period_minutes).total_seconds()
                + mlrun.mlconf.model_endpoint_monitoring.parquet_batching_timeout_secs
            )
        else:
            return True

    def run(self, event: nuclio_sdk.Event) -> None:
        """
        Main method for controller chief, runs all the relevant monitoring applications for a single endpoint.
        Handles nop events logic.
        This method handles the following:
        1. Read applications from the event (endpoint_policy)
        2. Check model monitoring windows
        3. Send data to applications
        4. Pushes nop event to main stream if needed
        """
        logger.info("Start running monitoring controller worker")
        try:
            body = json.loads(event.body.decode("utf-8"))
        except Exception as e:
            logger.error(
                "Failed to decode event",
                exc=err_to_str(e),
            )
            return
        # Run single endpoint process
        self.model_endpoint_process(event=body)

    def model_endpoint_process(
        self,
        event: Optional[dict] = None,
    ) -> None:
        """
        Process a model endpoint and trigger the monitoring applications. This function running on different process
        for each endpoint.

        :param event:                       (dict) Event that triggered the monitoring process.
        """
        logger.info("Model endpoint process started", event=event)

        try:
            project_name = event[ControllerEvent.PROJECT]
            endpoint_id = event[ControllerEvent.ENDPOINT_ID]
            not_old_batch_endpoint = True
            if (
                event[ControllerEvent.KIND]
                == mm_constants.ControllerEventKind.BATCH_COMPLETE
            ):
                monitoring_functions = (
                    self.project_obj.list_model_monitoring_functions()
                )
                if monitoring_functions:
                    applications_names = list(
                        {app.metadata.name for app in monitoring_functions}
                    )
                    last_stream_timestamp = datetime.datetime.fromisoformat(
                        event[ControllerEvent.LAST_TIMESTAMP]
                    )
                    first_request = datetime.datetime.fromisoformat(
                        event[ControllerEvent.FIRST_TIMESTAMP]
                    )
                    endpoint_mode = mm_constants.EndpointMode.BATCH
                    model_endpoint = self.project_obj.list_model_endpoints(
                        uids=[endpoint_id],
                        latest_only=True,
                    ).endpoints

                    if not model_endpoint:
                        logger.error(
                            "Batch model endpoint not found",
                            endpoint_id=endpoint_id,
                            project=project_name,
                        )
                        return

                    endpoint_name = model_endpoint[0].metadata.name
                    endpoint_updated = model_endpoint[0].metadata.updated.isoformat()

                else:
                    logger.info("No monitoring functions found", project=self.project)
                    return

            else:
                endpoint_name = event[ControllerEvent.ENDPOINT_NAME]
                applications_names = event[ControllerEvent.ENDPOINT_POLICY][
                    ControllerEventEndpointPolicy.MONITORING_APPLICATIONS
                ]
                last_stream_timestamp = datetime.datetime.fromisoformat(
                    event[ControllerEvent.TIMESTAMP]
                )
                first_request = datetime.datetime.fromisoformat(
                    event[ControllerEvent.FIRST_REQUEST]
                )

                endpoint_updated = event[ControllerEvent.ENDPOINT_POLICY][
                    ControllerEventEndpointPolicy.ENDPOINT_UPDATED
                ]

                endpoint_mode = mm_constants.EndpointMode.REAL_TIME

                not_old_batch_endpoint = (
                    event[ControllerEvent.ENDPOINT_TYPE] != EndpointType.BATCH_EP
                )

            logger.info(
                "Starting to analyze", timestamp=last_stream_timestamp.isoformat()
            )

            with _BatchWindowGenerator(
                project=project_name,
                endpoint_id=endpoint_id,
                window_length=self._window_length,
            ) as batch_window_generator:
                for application in applications_names:
                    for (
                        start_infer_time,
                        end_infer_time,
                    ) in batch_window_generator.get_intervals(
                        application=application,
                        first_request=first_request,
                        last_request=last_stream_timestamp,
                        endpoint_mode=endpoint_mode,
                        not_old_batch_endpoint=not_old_batch_endpoint,
                    ):
                        data_in_window = False
                        if not_old_batch_endpoint:
                            # Serving endpoint - get the relevant window data from the TSDB
                            prediction_metric = self.tsdb_connector.read_predictions(
                                start=start_infer_time,
                                end=end_infer_time,
                                endpoint_id=endpoint_id,
                            )
                            if prediction_metric.data:
                                data_in_window = True
                        else:
                            # Old batch endpoint - get the relevant window data from the parquet target
                            warnings.warn(
                                "Analyzing batch model endpoints with real time processing events is "
                                "deprecated in 1.10.0 and will be removed in 1.12.0. "
                                "Instead, use job-based serving to invoke and analyze offline batch model"
                                "endpoints.",
                                # TODO: Remove this in 1.12.0
                                FutureWarning,
                            )

                            if endpoint_id not in self.feature_sets:
                                self.feature_sets[endpoint_id] = fstore.get_feature_set(
                                    event[ControllerEvent.FEATURE_SET_URI]
                                )
                            self.feature_sets.move_to_end(endpoint_id, last=False)
                            if (
                                len(self.feature_sets)
                                > self._MAX_FEATURE_SET_PER_WORKER
                            ):
                                self.feature_sets.popitem(last=True)
                            m_fs = self.feature_sets.get(endpoint_id)

                            df = m_fs.to_dataframe(
                                start_time=start_infer_time,
                                end_time=end_infer_time,
                                time_column=mm_constants.EventFieldType.TIMESTAMP,
                                storage_options=self.storage_options,
                            )
                            if len(df) > 0:
                                data_in_window = True

                        if not data_in_window:
                            logger.info(
                                "No data found for the given interval",
                                start=start_infer_time,
                                end=end_infer_time,
                                endpoint_id=endpoint_id,
                            )
                        else:
                            logger.info(
                                "Data found for the given interval",
                                start=start_infer_time,
                                end=end_infer_time,
                                endpoint_id=endpoint_id,
                            )
                            self._push_to_applications(
                                start_infer_time=start_infer_time
                                - datetime.timedelta(
                                    batch_window_generator.batch_window.TIMESTAMP_RESOLUTION_MICRO
                                ),  # We subtract a microsecond to ensure that the apps will retrieve start time data.
                                end_infer_time=end_infer_time,
                                endpoint_id=endpoint_id,
                                endpoint_name=endpoint_name,
                                project=project_name,
                                applications_names=[application],
                                model_monitoring_access_key=self.model_monitoring_access_key,
                                endpoint_updated=endpoint_updated,
                            )

                if (
                    event[ControllerEvent.KIND]
                    == mm_constants.ControllerEventKind.REGULAR_EVENT
                ):
                    base_period = event[ControllerEvent.ENDPOINT_POLICY][
                        ControllerEventEndpointPolicy.BASE_PERIOD
                    ]
                    current_time = mlrun.utils.datetime_now()
                    if self._should_send_nop_event(
                        base_period,
                        batch_window_generator.get_min_last_analyzed(),
                        current_time,
                    ):
                        event = {
                            ControllerEvent.KIND: mm_constants.ControllerEventKind.NOP_EVENT,
                            ControllerEvent.PROJECT: project_name,
                            ControllerEvent.ENDPOINT_ID: endpoint_id,
                            ControllerEvent.ENDPOINT_NAME: endpoint_name,
                            ControllerEvent.TIMESTAMP: current_time.isoformat(
                                timespec="microseconds"
                            ),
                            ControllerEvent.ENDPOINT_POLICY: event[
                                ControllerEvent.ENDPOINT_POLICY
                            ],
                            ControllerEvent.ENDPOINT_TYPE: event[
                                ControllerEvent.ENDPOINT_TYPE
                            ],
                            ControllerEvent.FEATURE_SET_URI: event[
                                ControllerEvent.FEATURE_SET_URI
                            ],
                            ControllerEvent.FIRST_REQUEST: event[
                                ControllerEvent.FIRST_REQUEST
                            ],
                        }
                        self._push_to_main_stream(
                            event=event,
                            endpoint_id=endpoint_id,
                        )
            logger.info(
                "Finish analyze for",
                timestamp=last_stream_timestamp,
            )

        except Exception:
            logger.exception(
                "Encountered an exception",
                endpoint_id=event[ControllerEvent.ENDPOINT_ID],
            )

    def _push_to_applications(
        self,
        start_infer_time: datetime.datetime,
        end_infer_time: datetime.datetime,
        endpoint_id: str,
        endpoint_name: str,
        project: str,
        applications_names: list[str],
        model_monitoring_access_key: str,
        endpoint_updated: str,
    ):
        """
        Pushes data to multiple stream applications.

        :param start_infer_time:            The beginning of the infer interval window.
        :param end_infer_time:              The end of the infer interval window.
        :param endpoint_id:                 Identifier for the model endpoint.
        :param project: mlrun               Project name.
        :param applications_names:          List of application names to which data will be pushed.
        :param model_monitoring_access_key: Access key to apply the model monitoring process.
        :param endpoint_updated:            str isoformet for the timestamp the model endpoint was updated
        """
        data = {
            mm_constants.ApplicationEvent.START_INFER_TIME: start_infer_time.isoformat(
                sep=" ", timespec="microseconds"
            ),
            mm_constants.ApplicationEvent.END_INFER_TIME: end_infer_time.isoformat(
                sep=" ", timespec="microseconds"
            ),
            mm_constants.ApplicationEvent.ENDPOINT_ID: endpoint_id,
            mm_constants.ApplicationEvent.ENDPOINT_NAME: endpoint_name,
            mm_constants.ApplicationEvent.ENDPOINT_UPDATED: endpoint_updated,
        }
        for app_name in applications_names:
            data.update({mm_constants.ApplicationEvent.APPLICATION_NAME: app_name})
            if app_name not in self.applications_streams:
                self.applications_streams[app_name] = (
                    mlrun.model_monitoring.helpers.get_output_stream(
                        project=project,
                        function_name=app_name,
                        v3io_access_key=model_monitoring_access_key,
                    )
                )
            app_stream = self.applications_streams.get(app_name)

            logger.info(
                "Pushing data to application stream",
                endpoint_id=endpoint_id,
                app_name=app_name,
                app_stream_type=str(type(app_stream)),
            )
            app_stream.push([data], partition_key=endpoint_id)

    def push_regular_event_to_controller_stream(self) -> None:
        """
        pushes a regular event to the controller stream.
        """
        logger.info("Starting monitoring controller chief")
        applications_names = []
        endpoints = self.project_obj.list_model_endpoints(
            tsdb_metrics=False,
            modes=[
                mm_constants.EndpointMode.REAL_TIME,
                mm_constants.EndpointMode.BATCH_LEGACY,
            ],
        ).endpoints

        if not endpoints:
            logger.info("No model endpoints found", project=self.project)
            return

        last_request_dict = self.tsdb_connector.get_last_request(
            endpoint_ids=[mep.metadata.uid for mep in endpoints]
        )
        if isinstance(last_request_dict, pd.DataFrame):
            last_request_dict = last_request_dict.set_index(
                mm_constants.EventFieldType.ENDPOINT_ID
            )[mm_constants.ModelEndpointSchema.LAST_REQUEST].to_dict()

        monitoring_functions = self.project_obj.list_model_monitoring_functions()
        if monitoring_functions:
            # if monitoring_functions: - TODO : ML-7700
            #   Gets only application in ready state
            #   applications_names = list(
            #       {
            #           app.metadata.name
            #           for app in monitoring_functions
            #           if (
            #               app.status.state == "ready"
            #               # workaround for the default app, as its `status.state` is `None`
            #               or app.metadata.name
            #               == mm_constants.HistogramDataDriftApplicationConstants.NAME
            #           )
            #       }
            #   )
            applications_names = list(
                {app.metadata.name for app in monitoring_functions}
            )
        if not applications_names:
            logger.info("No monitoring functions found", project=self.project)
            return
        policy = {
            ControllerEventEndpointPolicy.MONITORING_APPLICATIONS: applications_names,
            ControllerEventEndpointPolicy.BASE_PERIOD: int(
                batch_dict2timedelta(
                    json.loads(
                        cast(
                            str,
                            os.getenv(mm_constants.EventFieldType.BATCH_INTERVALS_DICT),
                        )
                    )
                ).total_seconds()
                // _SECONDS_IN_MINUTE
            ),
        }
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=min(len(endpoints), 10)
        ) as pool:
            with schedules.ModelMonitoringSchedulesFileChief(
                self.project
            ) as schedule_file:
                for endpoint in endpoints:
                    last_request = last_request_dict.get(endpoint.metadata.uid, None)
                    if isinstance(last_request, float):
                        last_request = datetime.datetime.fromtimestamp(
                            last_request, tz=datetime.timezone.utc
                        )
                    elif isinstance(last_request, pd.Timestamp):
                        last_request = last_request.to_pydatetime()
                    endpoint.status.last_request = (
                        last_request or endpoint.status.last_request
                    )
                    futures = {
                        pool.submit(
                            self.endpoint_to_regular_event,
                            endpoint,
                            policy,
                            set(applications_names),
                            schedule_file,
                        ): endpoint
                    }
                for future in concurrent.futures.as_completed(futures):
                    if future.exception():
                        exception = future.exception()
                        error = (
                            f"Failed to push event. Endpoint name: {futures[future].metadata.name}, "
                            f"endpoint uid: {futures[future].metadata.uid}, traceback:\n"
                        )
                        error += "".join(
                            traceback.format_exception(
                                None, exception, exception.__traceback__
                            )
                        )
                        logger.error(error)
        logger.info("Finishing monitoring controller chief")

    def endpoint_to_regular_event(
        self,
        endpoint: mlrun.common.schemas.ModelEndpoint,
        policy: dict,
        applications_names: set,
        schedule_file: schedules.ModelMonitoringSchedulesFileChief,
    ) -> None:
        if self._should_monitor_endpoint(
            endpoint,
            set(applications_names),
            policy.get(ControllerEventEndpointPolicy.BASE_PERIOD, 10),
            schedule_file,
        ):
            logger.debug(
                "Endpoint data is being prepared for regular event",
                endpoint_id=endpoint.metadata.uid,
                endpoint_name=endpoint.metadata.name,
                timestamp=endpoint.status.last_request.isoformat(
                    sep=" ", timespec="microseconds"
                ),
                first_request=endpoint.status.first_request.isoformat(
                    sep=" ", timespec="microseconds"
                ),
                endpoint_type=endpoint.metadata.endpoint_type,
                feature_set_uri=endpoint.spec.monitoring_feature_set_uri,
                endpoint_policy=json.dumps(policy),
            )
            policy[ControllerEventEndpointPolicy.ENDPOINT_UPDATED] = (
                endpoint.metadata.updated.isoformat()
            )
            self.push_to_controller_stream(
                kind=mm_constants.ControllerEventKind.REGULAR_EVENT,
                project=endpoint.metadata.project,
                endpoint_id=endpoint.metadata.uid,
                endpoint_name=endpoint.metadata.name,
                timestamp=endpoint.status.last_request.isoformat(
                    sep=" ", timespec="microseconds"
                ),
                first_request=endpoint.status.first_request.isoformat(
                    sep=" ", timespec="microseconds"
                ),
                endpoint_type=endpoint.metadata.endpoint_type.value,
                feature_set_uri=endpoint.spec.monitoring_feature_set_uri,
                endpoint_policy=policy,
            )

    def push_to_controller_stream(
        self,
        kind: str,
        project: str,
        endpoint_id: str,
        endpoint_name: str,
        timestamp: str,
        first_request: str,
        endpoint_type: int,
        feature_set_uri: str,
        endpoint_policy: dict[str, Any],
    ) -> None:
        """
        Pushes event data to controller stream.
        :param timestamp: the event timestamp str isoformat utc timezone
        :param first_request: the first request str isoformat utc timezone
        :param endpoint_policy: dictionary hold the monitoring policy
        :param kind: str event kind
        :param project: project name
        :param endpoint_id: endpoint id string
        :param endpoint_name: the endpoint name string
        :param endpoint_type: Enum of the endpoint type
        :param feature_set_uri: the feature set uri string
        """
        event = {
            ControllerEvent.KIND.value: kind,
            ControllerEvent.PROJECT.value: project,
            ControllerEvent.ENDPOINT_ID.value: endpoint_id,
            ControllerEvent.ENDPOINT_NAME.value: endpoint_name,
            ControllerEvent.TIMESTAMP.value: timestamp,
            ControllerEvent.FIRST_REQUEST.value: first_request,
            ControllerEvent.ENDPOINT_TYPE.value: endpoint_type,
            ControllerEvent.FEATURE_SET_URI.value: feature_set_uri,
            ControllerEvent.ENDPOINT_POLICY.value: endpoint_policy,
        }
        logger.info(
            "Pushing data to controller stream",
            event=event,
            endpoint_id=endpoint_id,
            controller_stream_type=str(type(self.controller_stream)),
        )
        self.controller_stream.push([event], partition_key=endpoint_id)

    def _push_to_main_stream(self, event: dict, endpoint_id: str) -> None:
        """
        Pushes the given event to model monitoring stream
        :param event: event dictionary to push to stream
        :param endpoint_id: endpoint id string
        """
        logger.info(
            "Pushing data to main stream, NOP event is been generated",
            event=json.dumps(event),
            endpoint_id=endpoint_id,
            mm_stream_type=str(type(self.model_monitoring_stream)),
        )
        self.model_monitoring_stream.push([event], partition_key=endpoint_id)


def handler(context: nuclio_sdk.Context, event: nuclio_sdk.Event) -> None:
    """
    Run model monitoring application processor

    :param context: the Nuclio context
    :param event:   trigger event
    """
    logger.info(
        "Controller got event",
        trigger=event.trigger,
        trigger_kind=event.trigger.kind,
    )

    if event.trigger.kind in mm_constants.CRON_TRIGGER_KINDS:
        # Runs controller chief:
        context.user_data.monitor_app_controller.push_regular_event_to_controller_stream()
    elif event.trigger.kind in mm_constants.STREAM_TRIGGER_KINDS:
        # Runs controller worker:
        context.user_data.monitor_app_controller.run(event)
    else:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Wrong trigger kind for model monitoring controller"
        )


def init_context(context):
    monitor_app_controller = MonitoringApplicationController()
    setattr(context.user_data, "monitor_app_controller", monitor_app_controller)
    context.logger.info("Monitoring application controller initialized")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import abstractmethod
from datetime import datetime
from typing import Any, Callable, Generic, Optional, Union

import yaml

import mlrun
from mlrun.config import config

from ..utils import DLTypes
from .logger import Logger


class TensorboardLogger(Logger, Generic[DLTypes.WeightType]):
    """
    An abstract tensorboard logger class for logging the information collected during training / evaluation of the base
    logger to tensorboard. Each framework has its own way of logging to tensorboard, but each must implement the entire
    features listed in this class. The logging includes:

    * Summary text of the run with a hyperlink to the MLRun log if it was done.
    * Hyperparameters tuning table: static hyperparameters, dynamic hyperparameters and epoch validation summaries.
    * Plots:

      * Per iteration (batch) plot for the training and validation metrics.
      * Per epoch plot for the dynamic hyperparameters and validation summaries results.
      * Per epoch weights statistics for each weight and statistic.

    * Histograms per epoch for each of the logged weights.
    * Distributions per epoch for each of the logged weights.
    * Images per epoch for each of the logged weights.
    * Model architecture graph.
    """

    class _Sections:
        """
        Tensorboard split his plots to sections via a path like name <SECTION>/<PLOT_NAME>. These are the sections used
        in this callback for logging.
        """

        TRAINING = "Training"
        VALIDATION = "Validation"
        SUMMARY = "Summary"
        HYPERPARAMETERS = "Hyperparameters"
        WEIGHTS = "Weights"

    def __init__(
        self,
        statistics_functions: list[
            Callable[[DLTypes.WeightType], Union[float, DLTypes.WeightType]]
        ],
        context: mlrun.MLClientCtx = None,
        tensorboard_directory: Optional[str] = None,
        run_name: Optional[str] = None,
        update_frequency: Union[int, str] = "epoch",
    ):
        """
        Initialize a tensorboard logger callback with the given configuration. At least one of 'context' and
        'tensorboard_directory' must be given.

        :param statistics_functions:  A list of statistics functions to calculate at the end of each epoch on the
                                      tracked weights. Only relevant if weights are being tracked. The functions in
                                      the list must accept one Weight and return a float (or float convertible) value.
        :param context:               A MLRun context to use for logging into the user's tensorboard directory. The
                                      context parameters can be logged as static hyperparameters as well.
        :param tensorboard_directory: If context is not given, or if wished to set the directory even with context,
                                      this will be the output for the event logs of tensorboard. If not given, context
                                      must be provided as the default tensorboard output directory will be:
                                      /User/.tensorboard/<PROJECT_NAME> or if working on local, the set artifacts path.
        :param run_name:              This experiment run name. Each run name will be indexed at the end of the name so
                                      each experiment will be numbered automatically. If a context was given, the
                                      context's uid will be added instead of an index. If a run name was not given the
                                      current time stamp will be used.
        :param update_frequency:      Per how many iterations (batches) the callback should write the tracked values to
                                      tensorboard. Can be passed as a string equal to 'epoch' for per epoch and 'batch'
                                      for per single batch, or as an integer specifying per how many iterations to
                                      update. Notice that writing to tensorboard too frequently may cause the training
                                      to be slower. Default: 'epoch'.

        :raise MLRunInvalidArgumentError: If the `update_frequency` is illegal or if `tensorboard_directory` and
                                          `context` were not given.
        """
        super().__init__(context=context)

        # Validate the context and tensorboard directory combination:
        if tensorboard_directory is None and context is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "If context is not provided, the `tensorboard_directory` must be provided."
            )

        # Validate the update frequency:
        if isinstance(update_frequency, str) and update_frequency == "batch":
            update_frequency = 1
        elif not (
            (isinstance(update_frequency, str) and update_frequency == "epoch")
            or (isinstance(update_frequency, int) and update_frequency > 0)
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"The update frequency parameter is expected to be euqal to 'epoch', 'batch' or a positive "
                f"integer, received: {update_frequency}"
            )

        # Store the given parameters:
        self._statistics_functions = statistics_functions
        self._tensorboard_directory = tensorboard_directory
        self._run_name = run_name
        self._update_frequency = update_frequency

        # Setup the inner parameters:
        self._output_path = None
        self._last_logged_training_iteration = 0
        self._last_logged_validation_iteration = 0

        # Setup the weights dictionaries - a dictionary of all required weight parameters:
        # [Weight: str] -> [value: WeightType]
        self._weights = {}  # type: Dict[str, DLTypes.WeightType]

        # Setup the statistics dictionaries - a dictionary of statistics for the required weights per epoch:
        # [Statistic: str] -> [Weight: str] -> [epoch: int] -> [value: float]
        self._weights_statistics = {}  # type: Dict[str, Dict[str, List[float]]]
        for statistic_function in self._statistics_functions:
            self._weights_statistics[statistic_function.__name__] = {}  # type: Dict[str, List[float]]

    @property
    def weights(self) -> dict[str, DLTypes.WeightType]:
        """
        Get the logged weights dictionary. Each of the logged weight will be found by its name.

        :return: The weights dictionary.
        """
        return self._weights

    @property
    def weight_statistics(self) -> dict[str, dict[str, list[float]]]:
        """
        Get the logged statistics for all the tracked weights. Each statistic has a dictionary of weights and their list
        of epochs values.

        :return: The statistics dictionary.
        """
        return self._weights_statistics

    def log_weight(self, weight_name: str, weight_holder: DLTypes.WeightType):
        """
        Log the weight into the weights dictionary so it will be tracked and logged during the epochs. For each logged
        weight the key for it in the statistics logged will be initialized as well.

        :param weight_name:   The weight's name.
        :param weight_holder: The weight holder to track. Both Tensorflow (including Keras) and PyTorch (including
                              Lightning) keep the weights tensor in a holder object - 'Variable' for Tensorflow and
                              'Parameter' for PyTorch.
        """
        # Collect the given weight:
        self._weights[weight_name] = weight_holder

        # Insert the weight to all the statistics:
        for statistic in self._weights_statistics:
            self._weights_statistics[statistic][weight_name] = []

    def log_weights_statistics(self):
        """
        Calculate the statistics on the current weights and log the results.
        """
        for weight_name, weight_parameter in self._weights.items():
            for statistic_function in self._statistics_functions:
                self._weights_statistics[statistic_function.__name__][
                    weight_name
                ].append(float(statistic_function(weight_parameter)))

    def write_initial_summary_text(self):
        """
        Write the initial run summary text to tensorboard.
        """
        self._write_text_to_tensorboard(
            tag="MLRun", text=self._generate_run_start_text(), step=0
        )

    def write_epoch_summary_text(self):
        """
        Write the current epoch summary text to tensorboard.
        """
        self._write_text_to_tensorboard(
            tag="MLRun",
            text=self._generate_epoch_text(),
            step=self._training_iterations,
        )

    def write_final_summary_text(self):
        """
        Write the final training / evaluation run information to tensorboard.
        """
        # Log the run's final summary:
        self._write_text_to_tensorboard(
            tag="MLRun",
            text=self._generate_run_end_text(),
            step=(
                self._validation_iterations
                if self._training_iterations == 0
                else self._training_iterations
            ),
        )

    def write_training_results(self, ignore_update_frequency: bool = False) -> bool:
        """
        Write the current epoch training results with respect to the update frequency.

        :param ignore_update_frequency: If the update frequency should be ignored, meaning the training results will be
                                        written to tensorboard regardless the iteration. Should be happening at the end
                                        of every epoch.

        :return: True if the information was written to tensorboard and false if not.
        """
        # Check if the logger should write to tensorboard according to the update frequency:
        if not ignore_update_frequency and not self._write_to_tensorboard(
            is_training=True
        ):
            return False

        # Check if everything is already logged:
        if self._last_logged_training_iteration == self._training_iterations:
            return True

        # Get the index to log from:
        index_to_log_from = (
            self._training_iterations - self._last_logged_training_iteration
        )

        # Write the training results collected:
        for parameter, epochs in self._training_results.items():
            for i, value in enumerate(epochs[-1][-index_to_log_from:]):
                self._write_scalar_to_tensorboard(
                    name=f"{self._Sections.TRAINING}/{parameter}",
                    value=value,
                    step=self._last_logged_training_iteration + i + 1,
                )

        # Count the logged iterations:
        self._last_logged_training_iteration += index_to_log_from

        return True

    def write_validation_results(self, ignore_update_frequency: bool = False) -> bool:
        """
        Write the current epoch validation (evaluation) results with respect to the update frequency.

        :param ignore_update_frequency: If the update frequency should be ignored, meaning the training results will be
                                        written to tensorboard regardless the iteration. Should be happening at the end
                                        of every epoch.

        :return: True if the information was written to tensorboard and false if not.
        """
        # Check if the logger should write to tensorboard according to the update frequency:
        if not ignore_update_frequency and not self._write_to_tensorboard(
            is_training=False
        ):
            return False

        # Check if everything is already logged:
        if self._last_logged_validation_iteration == self._validation_iterations:
            return True

        # Get the index to log from:
        index_to_log_from = (
            self._validation_iterations - self._last_logged_validation_iteration
        )

        # Write the validation results collected:
        for parameter, epochs in self._validation_results.items():
            for i, value in enumerate(epochs[-1][-index_to_log_from:]):
                self._write_scalar_to_tensorboard(
                    name=f"{self._Sections.VALIDATION}/{parameter}",
                    value=value,
                    step=self._last_logged_validation_iteration + i + 1,
                )

        # Count the logged iterations:
        self._last_logged_validation_iteration += index_to_log_from

        return True

    def write_training_summaries(self):
        """
        Write the current epoch training metrics summaries values to tensorboard.
        """
        for metric, epochs in self._training_summaries.items():
            self._write_scalar_to_tensorboard(
                name=f"{self._Sections.SUMMARY}/training_{metric}",
                value=epochs[-1],
                step=self._epochs,
            )

    def write_validation_summaries(self):
        """
        Write the current epoch validation metrics summaries values to tensorboard.
        """
        for metric, epochs in self._validation_summaries.items():
            self._write_scalar_to_tensorboard(
                name=f"{self._Sections.SUMMARY}/validation_{metric}",
                value=epochs[-1],
                step=self._epochs,
            )

    def write_dynamic_hyperparameters(self):
        """
        Write the current epoch dynamic hyperparameter values to tensorboard.
        """
        for parameter, epochs in self._dynamic_hyperparameters.items():
            self._write_scalar_to_tensorboard(
                name=f"{self._Sections.HYPERPARAMETERS}/{parameter}",
                value=epochs[-1],
                step=self._epochs,
            )

    def write_weights_histograms(self):
        """
        Write the current epoch weights histograms to tensorboard.
        """
        for weight_name, weight in self._weights.items():
            self._write_weight_histogram_to_tensorboard(
                name=f"{self._Sections.WEIGHTS}/{weight_name}",
                weight=weight,
                step=self._epochs,
            )

    def write_weights_images(self):
        """
        Write the current epoch weights statistics to tensorboard.
        """
        for weight_name, weight in self._weights.items():
            self._write_weight_image_to_tensorboard(
                name=f"{self._Sections.WEIGHTS}/{weight_name}",
                weight=weight,
                step=self._epochs,
            )

    def write_weights_statistics(self):
        """
        Write the current epoch weight statistics to tensorboard.
        """
        for statistic, weights in self._weights_statistics.items():
            for weight_name, epoch_values in weights.items():
                self._write_scalar_to_tensorboard(
                    name=f"{self._Sections.WEIGHTS}/{weight_name}:{statistic}",
                    value=epoch_values[-1],
                    step=self._epochs,
                )

    @abstractmethod
    def write_model_to_tensorboard(self, *args, **kwargs):
        """
        Write the given model as a graph in tensorboard.
        """
        pass

    @abstractmethod
    def write_parameters_table_to_tensorboard(self):
        """
        Write the summaries, static and dynamic hyperparameters to the table in tensorboard's hparams section. This
        method is called once for creating the hparams table.
        """
        pass

    @abstractmethod
    def _write_text_to_tensorboard(self, tag: str, text: str, step: int):
        """
        Write text to tensorboard's text section. Summary information of this training / validation run will be logged
        to tensorboard using this method.

        :param tag:  The tag of the text (box it will be appearing under).
        :param text: The text to write.
        :param step: The iteration / epoch the text belongs to.
        """
        pass

    @abstractmethod
    def _write_scalar_to_tensorboard(self, name: str, value: float, step: int):
        """
        Write the scalar's value into its plot.

        :param name:  The plot's name.
        :param value: The value to add to the plot.
        :param step:  The iteration / epoch the value belongs to.
        """
        pass

    @abstractmethod
    def _write_weight_histogram_to_tensorboard(
        self, name: str, weight: DLTypes.WeightType, step: int
    ):
        """
        Write the current state of the weights as histograms to tensorboard.

        :param name:   The weight's name.
        :param weight: The weight to write its histogram.
        :param step:   The iteration / epoch the weight's histogram state belongs to.
        """
        pass

    @abstractmethod
    def _write_weight_image_to_tensorboard(
        self, name: str, weight: DLTypes.WeightType, step: int
    ):
        """
        Log the current state of the weights as images to tensorboard.

        :param name:   The weight's name.
        :param weight: The weight to write its image.
        :param step:   The iteration / epoch the weight's image state belongs to.
        """
        pass

    def _create_output_path(self):
        """
        Create the output path, indexing the given run name as needed. The output path will be:
        /self._tensorboard_directory/self._run_name_<index>.
        """
        # Set the default run name if it was not given:
        default_run_name = False
        if self._run_name is None:
            default_run_name = True
            self._run_name = (
                (
                    str(datetime.now())
                    .replace(" ", "_")
                    .replace(":", "-")
                    .replace(".", "-")
                )
                if self._context is None
                else f"{self._context.name}-{self._context.uid}"
            )

        # If the tensorboard directory is not provided, set it to the default:
        if self._tensorboard_directory is None:
            # Use the default tensorboard logs directory:
            self._tensorboard_directory = (
                mlrun.mlconf.default_tensorboard_logs_path.replace(
                    "{{project}}", self._context.project
                )
            )
            # Try to create the directory, if not succeeded (writing error) change to the artifacts path:
            try:
                os.makedirs(self._tensorboard_directory, exist_ok=True)
            except OSError:
                self._tensorboard_directory = self._context.artifact_path

        # Check if the run name needs to be indexed (if its not using context uid or time stamp):
        if not default_run_name:
            # Create the main tensorboard directory for searching in it:
            os.makedirs(self._tensorboard_directory, exist_ok=True)
            # Index the run name according to the tensorboard directory content:
            index = len(
                [
                    run_directory
                    for run_directory in sorted(os.listdir(self._tensorboard_directory))
                    if self._run_name == run_directory
                    or self._run_name == run_directory.rsplit("_", 1)[0]
                ]
            )
            # Check if need to index the name:
            if index > 0:
                self._run_name = f"{self._run_name}_{index + 1}"

        # Create the output path:
        self._output_path = os.path.join(self._tensorboard_directory, self._run_name)
        os.makedirs(self._output_path, exist_ok=True)

    def _write_to_tensorboard(self, is_training: bool) -> bool:
        """
        Whether or not the logger should write to tensorboard the specified information.

        :param is_training: Whether the information to update is of training or not (validation).

        :return: True if the logger should write to tensorboard according to the stored update frequency and False if
                 not.
        """
        if isinstance(self._update_frequency, str):
            return False
        if is_training:
            return self._training_iterations % self._update_frequency == 0
        return self._validation_iterations % self._update_frequency == 0

    def _generate_run_start_text(self) -> str:
        """
        Generate the run start summary text. The callback configuration will be written along side the context
        information - a hyperlink for the job in MLRun and the context metadata as strings.

        :return: The generated text.
        """
        # Write the main header:
        text = "###Run Properties"

        # Add the callbacks properties:
        text += "\n####Callback configuration:"
        for property_name, property_value in zip(
            ["Output directory", "Run name", "Tracked hyperparameters"],
            [
                self._output_path,
                self._run_name,
                list(self._static_hyperparameters.keys())
                + list(self._dynamic_hyperparameters.keys()),
            ],
        ):
            text += f"\n  * **{property_name.capitalize()}**: {self._markdown_print(value=property_value, tabs=2)}"

        # Add the context state:
        if self._context is not None:
            text += f"\n####Context initial state: ({self._generate_context_link(context=self._context)})"
            for property_name, property_value in self._extract_properties_from_context(
                context=self._context
            ).items():
                text += f"\n  * **{property_name.capitalize()}**: {self._markdown_print(value=property_value, tabs=2)}"

        return text

    def _generate_epoch_text(self) -> str:
        """
        Generate the last epoch summary text. If MLRun context is available, the results and artifacts will be
        displayed. Otherwise, the epochs results will be simply written.

        :return: The generated text.
        """
        text = f"####Epoch {self._epochs} summary:"
        if self._context is not None:
            for property_name, property_value in self._extract_properties_from_context(
                context=self._context
            ).items():
                if property_name not in [
                    "project",
                    "uid",
                    "state",
                    "name",
                    "labels",
                    "inputs",
                    "parameters",
                ]:
                    text += (
                        f"\n  * **{property_name.capitalize()}**: "
                        f"{self._markdown_print(value=property_value, tabs=2)}"
                    )
        else:
            for property_name, property_value in self._extract_epoch_results().items():
                text += f"\n  * **{property_name.capitalize()}**: {self._markdown_print(value=property_value, tabs=2)}"
        return text

    def _generate_run_end_text(self) -> str:
        """
        Generate the run end summary text, writing the final collected results and parameters values. If MLRun context
        is available the updated properties of the context will be written as well.

        :return: The generated text.
        """
        # Write the run summary:
        text = f"\n####Run final summary - epoch {self._epochs}:"
        for property_name, property_value in self._extract_epoch_results().items():
            text += f"\n  * **{property_name.capitalize()}**: {self._markdown_print(value=property_value, tabs=2)}"

        # Add the context final state:
        if self._context is not None:
            text += f"\n####Context final state: ({self._generate_context_link(context=self._context)})"
            for property_name, property_value in self._extract_properties_from_context(
                context=self._context
            ).items():
                text += f"\n  * **{property_name.capitalize()}**: {self._markdown_print(value=property_value, tabs=2)}"
        return text

    def _extract_epoch_results(
        self, epoch: int = -1
    ) -> dict[str, dict[str, DLTypes.TrackableType]]:
        """
        Extract the given epoch results from all the collected values and results.

        :param epoch: The epoch to get the results. Default: the last epoch (-1).

        :return: A dictionary where the keys are the collected value and the values are the results.
        """
        return {
            "Static hyperparameters": self._static_hyperparameters,
            "Dynamic hyperparameters": {
                name: value[epoch]
                for name, value in self._dynamic_hyperparameters.items()
            },
            "Training results": {
                name: value[epoch] for name, value in self._training_summaries.items()
            },
            "Validation results": {
                name: value[epoch] for name, value in self._validation_summaries.items()
            },
        }

    @staticmethod
    def _generate_context_link(
        context: mlrun.MLClientCtx, link_text: str = "view in MLRun"
    ) -> str:
        """
        Generate a hyperlink from the provided context to view in the MLRun web.

        :param context:   The context to generate his link.
        :param link_text: Text to present instead of the link.

        :return: The generated link.
        """
        return (
            f'<a href="{config.resolve_ui_url()}/{config.ui.projects_prefix}/{context.project}'
            f'/jobs/monitor/{context.uid}/overview" target="_blank">{link_text}</a>'
        )

    @staticmethod
    def _extract_properties_from_context(context: mlrun.MLClientCtx) -> dict[str, Any]:
        """
        Extract the properties of the run this context belongs to.

        :param context: The context to get his properties.

        :return: The properties as a dictionary where each key is the property name.
        """
        run = mlrun.RunObject.from_dict(context.to_dict())
        runs = mlrun.lists.RunList([run.to_dict()])
        info = {}
        for property_name, property_value in list(zip(*runs.to_rows())):
            info[property_name] = property_value
        return info

    @staticmethod
    def _markdown_print(value: Any, tabs: int = 0):
        """
        Convert the given value into a markdown styled string to print in tensorboard. List and dictionaries will both
        printed as yaml with minor differences.

        :param value: The value to print.
        :param tabs:  Indent to add at the beginning of every line.

        :return: The markdown styled string.
        """
        if isinstance(value, list):
            if len(value) == 0:
                return ""
            text = "\n" + yaml.safe_dump(value)
            text = "  \n".join(["  " * tabs + line for line in text.splitlines()])
            return text
        if isinstance(value, dict):
            if len(value) == 0:
                return ""
            text = yaml.safe_dump(value)
            text = "  \n".join(
                ["  " * tabs + "- " + line for line in text.splitlines()]
            )
            text = "  \n" + text
            return text
        return str(value)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""wasbfs - An fsspec driver for wasb/wasbs"""

__all__ = [
    "WasbFS",
]


import fsspec

from .fs import WasbFS

if hasattr(fsspec, "register_implementation"):
    # TODO: Not sure about clobber=True
    fsspec.register_implementation("wasbs", WasbFS, clobber=True)
    fsspec.register_implementation("wasb", WasbFS, clobber=True)
else:
    from fsspec.registry import known_implementations

    known_implementations["wasbs"] = {
        "class": "wasbfs.WasbFS",
        "err": "Please install WasbFS to use the wasb fileysstem class",
    }
    known_implementations["wasb"] = {
        "class": "wasbfs.WasbFS",
        "err": "Please install WasbFS to use the wasb fileysstem class",
    }

    del known_implementations

del fsspec  # clear the module namespace
